# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

import os
import signal
import subprocess
import sys
import time

from gi.repository import Gio
import dogtail.rawinput
import dogtail.tree

from virtinst import log
import tests.utils
from . import utils


class VMMDogtailApp:
    """
    Wrapper class to simplify dogtail app handling
    """

    def __init__(self, uri=tests.utils.URIs.test_full):
        self._proc = None
        self._root = None
        self._topwin = None
        self._manager = None
        self.uri = uri

    ####################################
    # Helpers to save testcase imports #
    ####################################

    def check(self, *args, **kwargs):
        return utils.check(*args, **kwargs)

    def sleep(self, *args, **kwargs):
        return time.sleep(*args, **kwargs)

    def find_window(self, name, roleName=None, check_active=True):
        if roleName is None:
            roleName = "(frame|dialog|alert|window)"
        return self.root.find(
            name=name, roleName=roleName, recursive=False, check_active=check_active
        )

    rawinput = dogtail.rawinput
    tree = dogtail.tree

    #################################
    # virt-manager specific helpers #
    #################################

    def get_manager(self, check_active=True):
        if not self._manager:
            self._manager = self.find_window("Virtual Machine Manager", check_active=check_active)
        return self._manager

    def find_details_window(self, vmname, click_details=False, shutdown=False):
        win = self.find_window("%s on" % vmname, "frame")
        if click_details:
            win.find("Details", "radio button").click()
        if shutdown:
            win.find("Shut Down", "push button").click()
            run = win.find("Run", "push button")
            utils.check(lambda: run.sensitive)
        return win

    def click_alert_button(self, label_text, button_text):
        alert = self.find_window("vmm dialog", "alert")
        alert.find_fuzzy(label_text, "label")
        alert.find(button_text, "push button").click()
        utils.check(lambda: not alert.active)

    def select_storagebrowser_volume(self, pool, vol, doubleclick=False):
        browsewin = self.find_window("vmm-storage-browser")
        browsewin.find_fuzzy(pool, "table cell").click()
        volcell = browsewin.find_fuzzy(vol, "table cell")
        if doubleclick:
            volcell.doubleClick()
        else:
            volcell.click()
            browsewin.find_fuzzy("Choose Volume").click()
        utils.check(lambda: not browsewin.active)

    ##########################
    # manager window helpers #
    ##########################

    def manager_open_createconn(self):
        manager = self.get_manager()
        manager.find("File", "menu").click()
        manager.find("Add Connection...", "menu item").click()
        win = self.root.find("Add Connection", "dialog")
        return win

    def manager_createconn(self, uri):
        win = self.manager_open_createconn()
        win.combo_select("Hypervisor", "Custom URI")
        win.find("uri-entry", "text").set_text(uri)
        win.find("Connect", "push button").click()
        utils.check(lambda: win.showing is False)

    def manager_get_conn_cell(self, conn_label):
        return self.get_manager().find(conn_label, "table cell")

    def manager_conn_connect(self, conn_label):
        c = self.manager_get_conn_cell(conn_label)
        c.click(button=3)
        self.root.find("conn-connect", "menu item").click()
        utils.check(lambda: "Not Connected" not in c.text)
        return c

    def manager_conn_disconnect(self, conn_label):
        c = self.manager_get_conn_cell(conn_label)
        c.click()
        utils.check(lambda: c.state_selected)
        c.click(button=3)
        menu = self.root.find("conn-menu", "menu")
        menu.find("conn-disconnect", "menu item").click()
        utils.check(lambda: "Not Connected" in c.text)
        return c

    def manager_conn_delete(self, conn_label):
        c = self.manager_get_conn_cell(conn_label)
        c.click(button=3)
        menu = self.root.find("conn-menu", "menu")
        menu.find("conn-delete", "menu item").click()
        self.click_alert_button("will remove the connection", "Yes")
        utils.check(lambda: c.dead)

    def manager_vm_action(
        self,
        vmname,
        confirm_click_no=False,
        run=False,
        shutdown=False,
        destroy=False,
        reset=False,
        reboot=False,
        pause=False,
        resume=False,
        save=False,
        restore=False,
        clone=False,
        migrate=False,
        delete=False,
        details=False,
    ):
        manager = self.get_manager()
        vmcell = manager.find(vmname + "\n", "table cell")

        if run:
            action = "Run"
        if shutdown:
            action = "Shut Down"
        if reboot:
            action = "Reboot"
        if reset:
            action = "Force Reset"
        if destroy:
            action = "Force Off"
        if pause:
            action = "Pause"
        if resume:
            action = "Resume"
        if save:
            action = "Save"
        if restore:
            action = "Restore"
        if clone:
            action = "Clone"
        if migrate:
            action = "Migrate"
        if delete:
            action = "Delete"
        if details:
            action = "Open"

        needs_shutdown = shutdown or destroy or reset or reboot or save
        needs_confirm = needs_shutdown or pause

        def _do_click():
            vmcell.click()
            vmcell.click(button=3)
            menu = self.root.find("vm-action-menu")
            utils.check(lambda: menu.onscreen)
            if needs_shutdown:
                smenu = menu.find("Shut Down", "menu")
                smenu.point()
                utils.check(lambda: smenu.onscreen)
                item = smenu.find(action, "menu item")
            else:
                item = menu.find(action, "menu item")
            utils.check(lambda: item.onscreen)
            item.point()
            utils.check(lambda: item.state_selected)
            item.click()
            return menu

        m = _do_click()
        if needs_confirm:
            if confirm_click_no:
                self.click_alert_button("Are you sure", "No")
                m = _do_click()
            self.click_alert_button("Are you sure", "Yes")
        utils.check(lambda: not m.onscreen)

    def manager_open_clone(self, vmname):
        self.manager_vm_action(vmname, clone=True)
        return self.find_window("Clone Virtual Machine")

    def manager_open_details(self, vmname, shutdown=False):
        self.manager_vm_action(vmname, details=True)
        win = self.find_details_window(vmname, shutdown=shutdown, click_details=True)
        return win

    def manager_open_host(self, tab, conn_label="test testdriver.xml"):
        """
        Helper to open host connection window and switch to a tab
        """
        self.root.find_fuzzy(conn_label, "table cell").click()
        self.root.find_fuzzy("Edit", "menu").click()
        self.root.find_fuzzy("Connection Details", "menu item").click()
        win = self.find_window("%s - Connection Details" % conn_label)
        tab = win.find_fuzzy(tab, "page tab")
        tab.point()
        tab.click()
        return win

    def manager_test_conn_window_cleanup(self, conn_label, childwin):
        # Give time for the child window to appear and possibly grab focus
        self.sleep(1)
        self.get_manager(check_active=False)
        dogtail.rawinput.dragWithTrajectory(childwin.title_coordinates(), (1000, 1000))
        self.manager_conn_disconnect(conn_label)
        utils.check(lambda: not childwin.showing)

    ###########################
    # Process management APIs #
    ###########################

    @property
    def root(self):
        if self._root is None:
            self.open()
        return self._root

    @property
    def topwin(self):
        if self._topwin is None:
            self.open()
        return self._topwin

    def has_dbus(self):
        dbus = Gio.DBusProxy.new_sync(
            Gio.bus_get_sync(Gio.BusType.SESSION, None),
            0,
            None,
            "org.freedesktop.DBus",
            "/org/freedesktop/DBus",
            "org.freedesktop.DBus",
            None,
        )
        return "org.virt-manager.virt-manager" in dbus.ListNames()

    def error_if_already_running(self):
        # Ensure virt-manager isn't already running
        if self.has_dbus():
            raise RuntimeError(
                "virt-manager is already running. Close it before running this test suite."
            )

    def is_running(self):
        return bool(self._proc and self._proc.poll() is None)

    def wait_for_exit(self):
        # Wait for shutdown for 2 sec
        waittime = 5
        self._proc.wait(timeout=waittime)

    def stop(self):
        """
        Try graceful process shutdown, then kill it
        """
        if not self._proc:
            return

        try:
            self._proc.send_signal(signal.SIGINT)
        except Exception:
            log.debug("Error terminating process", exc_info=True)
            self._proc = None
            return

        try:
            self.wait_for_exit()
        except subprocess.TimeoutExpired:
            log.warning("App didn't exit gracefully from SIGINT. Killing...")
            self._proc.kill()
            self.wait_for_exit()
            raise

    #####################################
    # virt-manager launching entrypoint #
    #####################################

    def open(
        self,
        uri=None,
        extra_opts=None,
        check_already_running=True,
        use_uri=True,
        window_name=None,
        xmleditor_enabled=False,
        keyfile=None,
        break_setfacl=False,
        first_run=True,
        will_fail=False,
        enable_libguestfs=False,
        firstrun_uri=None,
        show_console=None,
        allow_debug=True,
    ):
        extra_opts = extra_opts or []
        uri = uri or self.uri

        if allow_debug and tests.utils.TESTCONFIG.debug:
            stdout = sys.stdout
            stderr = sys.stderr
            extra_opts.append("--debug")
        else:
            stdout = open(os.devnull)
            stderr = open(os.devnull)

        cmd = [sys.executable]
        cmd += [os.path.join(tests.utils.TOPDIR, "virt-manager")]
        if use_uri:
            cmd += ["--connect", uri]
        if show_console:
            cmd += ["--show-domain-console=%s" % show_console]

        if first_run:
            cmd.append("--test-options=first-run")
            if not firstrun_uri:
                firstrun_uri = ""
        if firstrun_uri is not None:
            cmd.append("--test-options=firstrun-uri=%s" % firstrun_uri)
        if xmleditor_enabled:
            cmd.append("--test-options=xmleditor-enabled")
        if break_setfacl:
            cmd.append("--test-options=break-setfacl")
        if enable_libguestfs is True:
            cmd.append("--test-options=enable-libguestfs")
        if enable_libguestfs is False:
            cmd.append("--test-options=disable-libguestfs")
        if keyfile:
            import atexit
            import tempfile

            keyfile = tests.utils.UITESTDATADIR + "/keyfile/" + keyfile
            tempname = tempfile.mktemp(prefix="virtmanager-uitests-keyfile")
            open(tempname, "w").write(open(keyfile).read())
            atexit.register(lambda: os.unlink(tempname))
            cmd.append("--test-options=gsettings-keyfile=%s" % tempname)

        cmd += extra_opts

        if check_already_running:
            self.error_if_already_running()
        self._proc = subprocess.Popen(cmd, stdout=stdout, stderr=stderr)
        if will_fail:
            return

        with utils.dogtail_timeout(10):
            # On Fedora 39 sometimes app launch from the test suite
            # takes a while for reasons I can't quite figure
            self._root = dogtail.tree.root.application("virt-manager")
            self._topwin = self.find_window(window_name)



# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

import os
import signal
import sys
import warnings

# Dogtail is noisy with GTK and GI deprecation warnings
warnings.simplefilter("ignore")

# Ignores pylint error since dogtail doesn't specify this
import gi

gi.require_version("Atspi", "2.0")

import dogtail.config
import dogtail.utils

# find() backoff handling
dogtail.config.config.searchBackoffDuration = 0.1
dogtail.config.config.searchCutoffCount = 20

# Use .1 second delay between each action (default 1)
dogtail.config.config.actionDelay = 0.1
dogtail.config.config.defaultDelay = 0.1

# Turn off needlessly noisy debugging
DOGTAIL_DEBUG = False
dogtail.config.config.logDebugToStdOut = DOGTAIL_DEBUG
dogtail.config.config.logDebugToFile = False

# Dogtail screws with the default excepthook, disabling output if we turned
# off logging, so fix it
sys.excepthook = sys.__excepthook__

# dogtail.utils.Blinker creates a GLib.MainLoop on module import, which
# screws up SIGINT handling somehow. This reregisters the
# unittest.installHandler magic
signal.signal(signal.SIGINT, signal.getsignal(signal.SIGINT))

os.environ.pop("VIRTINST_TEST_SUITE", None)

if not dogtail.utils.isA11yEnabled():
    print("Enabling gsettings accessibility")
    dogtail.utils.enableA11y()

# This will trigger an error if accessibility isn't enabled
import dogtail.tree  # pylint: disable=wrong-import-order,ungrouped-imports



#
# Copyright 2013 Red Hat, Inc.
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.
"""
Classes for building and installing libvirt <network> XML
"""

from .xmlbuilder import XMLBuilder, XMLChildProperty, XMLProperty


class _NetworkDHCPRange(XMLBuilder):
    XML_NAME = "range"
    start = XMLProperty("./@start")
    end = XMLProperty("./@end")


class _NetworkDHCPHost(XMLBuilder):
    XML_NAME = "host"
    macaddr = XMLProperty("./@mac")
    name = XMLProperty("./@name")
    ip = XMLProperty("./@ip")


class _NetworkIP(XMLBuilder):
    XML_NAME = "ip"

    family = XMLProperty("./@family")
    address = XMLProperty("./@address")
    prefix = XMLProperty("./@prefix", is_int=True)
    netmask = XMLProperty("./@netmask")

    tftp = XMLProperty("./tftp/@root")
    bootp_file = XMLProperty("./dhcp/bootp/@file")
    bootp_server = XMLProperty("./dhcp/bootp/@server")

    ranges = XMLChildProperty(_NetworkDHCPRange, relative_xpath="./dhcp")
    hosts = XMLChildProperty(_NetworkDHCPHost, relative_xpath="./dhcp")


class _NetworkRoute(XMLBuilder):
    XML_NAME = "route"

    family = XMLProperty("./@family")
    address = XMLProperty("./@address")
    prefix = XMLProperty("./@prefix", is_int=True)
    gateway = XMLProperty("./@gateway")
    netmask = XMLProperty("./@netmask")


class _NetworkForwardPf(XMLBuilder):
    XML_NAME = "pf"
    dev = XMLProperty("./@dev")


class _NetworkForward(XMLBuilder):
    XML_NAME = "forward"

    mode = XMLProperty("./@mode")
    dev = XMLProperty("./@dev")
    managed = XMLProperty("./@managed")
    pf = XMLChildProperty(_NetworkForwardPf)


class _NetworkPortgroup(XMLBuilder):
    XML_NAME = "portgroup"

    name = XMLProperty("./@name")
    default = XMLProperty("./@default", is_yesno=True)


class Network(XMLBuilder):
    """
    Top level class for <network> object XML
    """

    XML_NAME = "network"
    _XML_PROP_ORDER = [
        "ipv6",
        "name",
        "uuid",
        "forward",
        "virtualport_type",
        "bridge",
        "stp",
        "delay",
        "domain_name",
        "macaddr",
        "ips",
        "routes",
    ]

    ipv6 = XMLProperty("./@ipv6", is_yesno=True)
    name = XMLProperty("./name")
    uuid = XMLProperty("./uuid")

    virtualport_type = XMLProperty("./virtualport/@type")

    # Not entirely correct, there can be multiple routes
    forward = XMLChildProperty(_NetworkForward, is_single=True)

    domain_name = XMLProperty("./domain/@name")

    bridge = XMLProperty("./bridge/@name")
    stp = XMLProperty("./bridge/@stp", is_onoff=True)
    delay = XMLProperty("./bridge/@delay", is_int=True)
    macaddr = XMLProperty("./mac/@address")

    portgroups = XMLChildProperty(_NetworkPortgroup)
    ips = XMLChildProperty(_NetworkIP)
    routes = XMLChildProperty(_NetworkRoute)

    ###################
    # Helper routines #
    ###################

    def can_pxe(self):
        forward = self.forward.mode
        if forward and forward != "nat":
            return True
        for ip in self.ips:
            if ip.bootp_file:
                return True
        return False



# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

from . import lib


######################################
# UI tests for the createpool wizard #
######################################


def _open_createpool(app, hostwin):
    hostwin.find("pool-add", "push button").click()
    win = app.find_window("Add a New Storage Pool")
    lib.utils.check(lambda: win.active)
    return win


def testCreatePools(app):
    hostwin = app.manager_open_host("Storage")
    win = _open_createpool(app, hostwin)
    finish = win.find("Finish", "push button")
    name = win.find("Name:", "text")

    def _browse_local_path(winlabel, usepath):
        chooser = app.root.find(winlabel, "file chooser")
        # Enter the filename and select it
        chooser.find(usepath, "table cell").click()
        obutton = chooser.find("Open", "push button")
        lib.utils.check(lambda: obutton.sensitive)
        obutton.click()
        lib.utils.check(lambda: not chooser.showing)
        lib.utils.check(lambda: win.active)

    # Create a simple default dir pool
    lib.utils.check(lambda: name.text == "pool")
    newname = "a-test-new-pool"
    name.set_text(newname)
    finish.click()

    # Select the new object in the host window, then do
    # stop->start->stop->delete, for lifecycle testing
    lib.utils.check(lambda: hostwin.active)
    cell = hostwin.find(newname, "table cell")
    delete = hostwin.find("pool-delete", "push button")
    start = hostwin.find("pool-start", "push button")
    stop = hostwin.find("pool-stop", "push button")

    cell.click()
    stop.click()
    lib.utils.check(lambda: start.sensitive)
    start.click()
    lib.utils.check(lambda: stop.sensitive)
    stop.click()
    lib.utils.check(lambda: delete.sensitive)

    # Delete it, clicking 'No' first
    delete.click()
    app.click_alert_button("permanently delete the pool", "No")
    lib.utils.check(lambda: not cell.dead)
    delete.click()
    app.click_alert_button("permanently delete the pool", "Yes")
    # Ensure it's gone
    lib.utils.check(lambda: cell.dead)

    # Test a disk pool
    win = _open_createpool(app, hostwin)
    win.combo_select("Type:", "disk:")
    newname = "a-pool-logical"
    name.set_text("a-pool-logical")
    win.find("source-browse").click()
    _browse_local_path("Choose source path", "console")
    finish.click()
    hostwin.find(newname, "table cell")

    # Test a iscsi pool
    win = _open_createpool(app, hostwin)
    win.combo_select("Type:", "iscsi:")
    newname = "a-pool-iscsi"
    name.set_text("a-pool-iscsi")
    win.find("target-browse").click()
    _browse_local_path("Choose target directory", "by-path")
    finish.click()
    # Catch example error
    app.click_alert_button("source host name", "Close")
    win.find("Host Name:", "text").set_text("example.com")
    win.find("pool-source-path-text").set_text("foo-iqn")
    win.find_fuzzy("Initiator IQN:", "check").click()
    win.find("iqn-text", "text").set_text("initiator-foo")
    finish.click()
    hostwin.find(newname, "table cell")

    # Test a logical pool
    win = _open_createpool(app, hostwin)
    win.combo_select("Type:", "logical:")
    newname = "a-lvm-pool"
    name.set_text("a-lvm-pool")

    win.combo_check_default("Volgroup", "testvg1")
    win.combo_select("Volgroup", "testvg2")
    finish.click()
    hostwin.find(newname, "table cell")

    # Test a scsi pool
    win = _open_createpool(app, hostwin)
    win.combo_select("Type:", "scsi:")
    newname = "a-scsi-pool"
    name.set_text("a-scsi-pool")
    win.combo_select("Source Adapter:", "host2")
    finish.click()
    hostwin.find(newname, "table cell")

    # Test a ceph pool
    win = _open_createpool(app, hostwin)
    newname = "a-ceph-pool"
    name.set_text("a-ceph-pool")
    win.combo_select("Type:", "rbd:")
    win.find_fuzzy("Host Name:", "text").set_text("example.com:1234")
    win.find_fuzzy("pool-source-name-text", "text").typeText("frob")
    finish.click()
    lib.utils.check(lambda: not win.showing)
    lib.utils.check(lambda: hostwin.active)
    hostwin.find(newname, "table cell")

    # Ensure host window closes fine
    hostwin.click()
    hostwin.keyCombo("<ctrl>w")
    lib.utils.check(lambda: not hostwin.showing)


def testCreatePoolXMLEditor(app):
    app.open(xmleditor_enabled=True)
    hostwin = app.manager_open_host("Storage")
    win = _open_createpool(app, hostwin)
    finish = win.find("Finish", "push button")
    name = win.find("Name:", "text")

    # Create a new obj with XML edited name, verify it worked
    tmpname = "objtmpname"
    newname = "froofroo"
    name.set_text(tmpname)
    win.find("XML", "page tab").click()
    xmleditor = win.find("XML editor")
    newtext = xmleditor.text.replace(">%s<" % tmpname, ">%s<" % newname)
    xmleditor.set_text(newtext)
    finish.click()
    lib.utils.check(lambda: hostwin.active)
    cell = hostwin.find(newname, "table cell")
    cell.click()

    # Do standard xmleditor tests
    win = _open_createpool(app, hostwin)
    lib.utils.test_xmleditor_interactions(app, win, finish)
    win.find("Cancel", "push button").click()
    lib.utils.check(lambda: not win.visible)



# Copyright (C) 2013, 2014 Red Hat, Inc.
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

import importlib
import os

# Need to do this before any tests or virtinst import
os.environ["VIRTINST_TEST_SUITE"] = "1"
# Need to do this before we import argcomplete
os.environ.pop("_ARC_DEBUG", None)
# Make sure the test suite uses an English locale, as we need to match
# error/status messages
os.environ["LANG"] = "en_US.UTF-8"
os.environ.pop("LANGUAGE", None)

# pylint: disable=wrong-import-position
from virtinst import buildconfig
from virtinst import log, reset_logging

# This sets all the cli bits back to their defaults
importlib.reload(buildconfig)

from tests import utils

# pylint: disable=ungrouped-imports
from virtinst import virtinstall
from virtinst import virtclone
from virtinst import virtxml


def setup_logging():
    import logging

    reset_logging()

    fmt = "%(levelname)-8s %(message)s"
    streamHandler = logging.StreamHandler()
    streamHandler.setFormatter(logging.Formatter(fmt))
    if utils.TESTCONFIG.debug:
        streamHandler.setLevel(logging.DEBUG)
        log.setLevel(logging.DEBUG)
    else:
        streamHandler.setLevel(logging.ERROR)
        log.setLevel(logging.ERROR)
    log.addHandler(streamHandler)



# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

import pytest


@pytest.fixture
def app():
    """
    Custom pytest fixture to a VMMDogtailApp instance to the testcase
    """
    from .lib.app import VMMDogtailApp

    testapp = VMMDogtailApp()
    try:
        yield testapp
    finally:
        testapp.stop()



# Copyright 2019 Red Hat, Inc.
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

import os
import random
import re
import string
import tempfile

from ..logger import log
from .. import progress
from . import urlfetcher


class _CloudInitConfig:
    """
    Helper class to create cloud-init configuration file.

    @destfile: Name of the cloud-init configuration file to be included
        in cloud-init ISO file.
    @content: Function to create the content of @destfile. It should return str
        to create the configuration file or None to not create the file.
    @config: Path to configuration file user provided to virt-install, if provided
        @content is ignored.
    @scratchdir: Directory where to place temporary files.
    """

    def __init__(self, destfile, content, config, scratchdir):
        self._destfile = destfile
        self._content = content
        self._config = config
        self._scratchdir = scratchdir

    def _create_file(self):
        content = self._content()

        if not content:
            return None

        fileobj = tempfile.NamedTemporaryFile(
            prefix="virtinst-", suffix=f"-{self._destfile}", dir=self._scratchdir, delete=False
        )
        with open(fileobj.name, "w+") as f:
            f.write(content)
        return fileobj.name

    def _fetch_file(self):
        log.debug("Using '%s' content from path='%s'", self._destfile, self._config)
        meter = progress.make_meter(quiet=True)
        fetcher = urlfetcher.DirectFetcher(None, self._scratchdir, meter)
        return fetcher.acquireFile(self._config)

    def add_filepair(self, filepairs):
        if self._config:
            file = self._fetch_file()
        else:
            file = self._create_file()

        if not file:
            return

        filepairs.append((file, self._destfile))


class CloudInitData:
    def __init__(self):
        self.disable = None
        self.root_password_generate = None
        self.root_password_file = None
        self.generated_root_password = None
        self.root_ssh_key = None
        self.clouduser_ssh_key = None
        self.user_data = None
        self.meta_data = None
        self.network_config = None

    def _generate_password(self):
        if not self.generated_root_password:
            self.generated_root_password = ""
            for dummy in range(16):
                self.generated_root_password += random.choice(string.ascii_letters + string.digits)
        return self.generated_root_password

    def _get_password(self, pwdfile):
        with open(pwdfile, "r") as fobj:
            return fobj.readline().rstrip("\n\r")

    def get_password_if_generated(self):
        if self.root_password_generate:
            return self._generate_password()

    def get_root_password(self):
        if self.root_password_file:
            return self._get_password(self.root_password_file)
        return self.get_password_if_generated()

    def get_root_ssh_key(self):
        if self.root_ssh_key:
            return self._get_password(self.root_ssh_key)

    def get_clouduser_ssh_key(self):
        if self.clouduser_ssh_key:
            return self._get_password(self.clouduser_ssh_key)

    def _create_metadata_content(self):
        return ""

    def _create_userdata_content(self):
        content = "#cloud-config\n"

        if self.root_password_generate or self.root_password_file:
            rootpass = self.get_root_password()
            content += "chpasswd:\n"
            content += "  list: |\n"
            content += "    root:%s\n" % rootpass

        if self.root_password_generate:
            content += "  expire: True\n"
        elif self.root_password_file:
            content += "  expire: False\n"

        if self.root_ssh_key:
            rootkey = self.get_root_ssh_key()
            content += "users:\n"
            content += "  - default\n"
            content += "  - name: root\n"
            content += "    ssh_authorized_keys:\n"
            content += "      - %s\n" % rootkey

        if self.clouduser_ssh_key:
            userkey = self.get_clouduser_ssh_key()
            content += "ssh_authorized_keys:\n"
            content += "  - %s\n" % userkey

        if self.disable:
            content += "runcmd:\n"
            content += '- echo "Disabled by virt-install" > /etc/cloud/cloud-init.disabled\n'

        clean_content = re.sub(r"root:(.*)", "root:[SCRUBBLED]", content)
        if "VIRTINST_TEST_SUITE_PRINT_CLOUDINIT" in os.environ:
            print(clean_content)

        log.debug("Generated cloud-init userdata: \n%s", clean_content)
        return content

    def _create_network_config_content(self):
        return None

    def create_files(self, scratchdir):
        meta_data = _CloudInitConfig(
            "meta-data", self._create_metadata_content, self.meta_data, scratchdir
        )
        user_data = _CloudInitConfig(
            "user-data", self._create_userdata_content, self.user_data, scratchdir
        )
        network_config = _CloudInitConfig(
            "network-config", self._create_network_config_content, self.network_config, scratchdir
        )

        filepairs = []
        try:
            meta_data.add_filepair(filepairs)
            user_data.add_filepair(filepairs)
            network_config.add_filepair(filepairs)
        except Exception:  # pragma: no cover
            for filepair in filepairs:
                os.unlink(filepair[0])
            raise

        return filepairs



# Copyright (C) 2013 Red Hat, Inc.
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

import os

from virtinst import StoragePool, StorageVolume
from virtinst import log

from tests import utils

# pylint: disable=protected-access

BASEPATH = os.path.join(utils.DATADIR, "storage")


def createPool(
    conn,
    ptype,
    poolname=None,
    fmt=None,
    target_path=None,
    source_path=None,
    source_name=None,
    iqn=None,
):

    if poolname is None:
        poolname = StoragePool.find_free_name(conn, "%s-pool" % ptype)

    pool_inst = StoragePool(conn)
    pool_inst.name = poolname
    pool_inst.type = ptype

    if pool_inst.supports_hosts():
        hostobj = pool_inst.hosts.add_new()
        hostobj.name = "some.random.hostname"
    if pool_inst.supports_source_path():
        pool_inst.source_path = source_path or "/some/source/path"
    if pool_inst.supports_target_path():
        pool_inst.target_path = target_path or pool_inst.default_target_path()
    if fmt and pool_inst.supports_format():
        pool_inst.format = fmt
    if pool_inst.supports_source_name():
        pool_inst.source_name = source_name or pool_inst.default_source_name()
    if iqn and pool_inst.supports_iqn():
        pool_inst.iqn = iqn

    return poolCompare(pool_inst)


def removePool(poolobj):
    poolobj.destroy()
    poolobj.undefine()


def poolCompare(pool_inst):
    pool_inst.validate()
    filename = os.path.join(BASEPATH, pool_inst.name + ".xml")
    out_expect = pool_inst.get_xml()

    if not os.path.exists(filename):
        open(filename, "w").write(out_expect)
    utils.diff_compare(out_expect, filename)

    return pool_inst.install(build=True, meter=None, create=True)


def createVol(conn, poolobj, volname=None, input_vol=None, clone_vol=None):
    if volname is None:
        volname = poolobj.name() + "-vol"

    alloc = 5 * 1024 * 1024 * 1024
    cap = 10 * 1024 * 1024 * 1024
    vol_inst = StorageVolume(conn)
    vol_inst.pool = poolobj
    vol_inst.name = volname
    vol_inst.capacity = cap
    vol_inst.allocation = alloc

    vol_inst.permissions.mode = "0700"
    vol_inst.permissions.owner = "10736"
    vol_inst.permissions.group = "10736"

    if input_vol:
        vol_inst.set_input_vol(input_vol)
    elif clone_vol:
        vol_inst = StorageVolume(conn, parsexml=clone_vol.XMLDesc(0))
        vol_inst.set_input_vol(clone_vol)
        vol_inst.name = volname

    vol_inst.validate()
    filename = os.path.join(BASEPATH, vol_inst.name + ".xml")

    # Format here depends on libvirt-7.2.0 and later
    if clone_vol and conn.local_libvirt_version() < 7002000:
        log.debug("skip clone compare")
        return

    utils.diff_compare(vol_inst.get_xml(), filename)
    return vol_inst.install(meter=False)


##############
# Test cases #
##############


def testDirPool():
    conn = utils.URIs.open_testdefault_cached()
    poolobj = createPool(conn, StoragePool.TYPE_DIR, "pool-dir2")
    invol = createVol(conn, poolobj)
    createVol(conn, poolobj, volname=invol.name() + "input", input_vol=invol)
    createVol(conn, poolobj, volname=invol.name() + "clone", clone_vol=invol)
    removePool(poolobj)


def testFSPool():
    conn = utils.URIs.open_testdefault_cached()
    poolobj = createPool(conn, StoragePool.TYPE_FS, "pool-fs")
    invol = createVol(conn, poolobj)
    createVol(conn, poolobj, volname=invol.name() + "input", input_vol=invol)
    createVol(conn, poolobj, volname=invol.name() + "clone", clone_vol=invol)
    removePool(poolobj)


def testNetFSPool():
    conn = utils.URIs.open_testdefault_cached()
    poolobj = createPool(conn, StoragePool.TYPE_NETFS, "pool-netfs")
    invol = createVol(conn, poolobj)
    createVol(conn, poolobj, volname=invol.name() + "input", input_vol=invol)
    createVol(conn, poolobj, volname=invol.name() + "clone", clone_vol=invol)
    removePool(poolobj)


def testLVPool():
    conn = utils.URIs.open_testdefault_cached()
    poolobj = createPool(conn, StoragePool.TYPE_LOGICAL, "pool-logical", source_name="pool-logical")
    invol = createVol(conn, poolobj)
    createVol(conn, poolobj, volname=invol.name() + "input", input_vol=invol)
    createVol(conn, poolobj, volname=invol.name() + "clone", clone_vol=invol)
    removePool(poolobj)

    # Test parsing source name for target path
    poolobj = createPool(
        conn, StoragePool.TYPE_LOGICAL, "pool-logical-target-srcname", target_path="/dev/vgfoobar"
    )
    removePool(poolobj)

    # Test with source name
    poolobj = createPool(
        conn, StoragePool.TYPE_LOGICAL, "pool-logical-srcname", source_name="vgname"
    )
    removePool(poolobj)


def testDiskPool():
    conn = utils.URIs.open_testdefault_cached()
    poolobj = createPool(
        conn, StoragePool.TYPE_DISK, "pool-disk", fmt="auto", target_path="/some/target/path"
    )
    invol = createVol(conn, poolobj)
    createVol(conn, poolobj, volname=invol.name() + "input", input_vol=invol)
    createVol(conn, poolobj, volname=invol.name() + "clone", clone_vol=invol)
    removePool(poolobj)


def testISCSIPool():
    conn = utils.URIs.open_testdefault_cached()
    poolobj = createPool(conn, StoragePool.TYPE_ISCSI, "pool-iscsi", iqn="foo.bar.baz.iqn")
    removePool(poolobj)


def testSCSIPool():
    conn = utils.URIs.open_testdefault_cached()
    poolobj = createPool(conn, StoragePool.TYPE_SCSI, "pool-scsi")
    removePool(poolobj)


def testMpathPool():
    conn = utils.URIs.open_testdefault_cached()
    poolobj = createPool(conn, StoragePool.TYPE_MPATH, "pool-mpath")
    removePool(poolobj)


def testGlusterPool():
    conn = utils.URIs.open_testdefault_cached()
    poolobj = createPool(conn, StoragePool.TYPE_GLUSTER, "pool-gluster")
    removePool(poolobj)


def testRBDPool():
    conn = utils.URIs.open_testdefault_cached()
    poolobj = createPool(conn, StoragePool.TYPE_RBD, "pool-rbd")
    removePool(poolobj)


def testMisc():
    conn = utils.URIs.open_testdefault_cached()
    # Misc coverage testing
    vol = StorageVolume(conn)
    assert vol.is_size_conflict()[0] is False

    fullconn = utils.URIs.open_testdriver_cached()
    glusterpool = fullconn.storagePoolLookupByName("pool-gluster")
    diskpool = fullconn.storagePoolLookupByName("pool-logical")

    glustervol = StorageVolume(fullconn)
    glustervol.pool = glusterpool
    assert glustervol.supports_format() is False

    diskvol = StorageVolume(fullconn)
    diskvol.pool = diskpool
    assert diskvol.supports_format() is False

    glusterpool.destroy()
    StoragePool.ensure_pool_is_running(glusterpool)

    # Check pool collision detection
    name = StoragePool.find_free_name(fullconn, "pool-gluster")
    assert name == "pool-gluster-1"


def testEnumerateLogical():
    conn = utils.URIs.open_testdefault_cached()
    lst = StoragePool.pool_list_from_sources(conn, StoragePool.TYPE_LOGICAL)
    assert lst == ["testvg1", "testvg2"]



# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

import os

import pytest

import tests
from tests.utils import TESTCONFIG


def pytest_addoption(parser):
    parser.addoption("--uitests", action="store_true", default=False, help="Run dogtail UI tests")

    parser.addoption(
        "--regenerate-output", action="store_true", default=False, help="Regenerate test output"
    )

    # test_urls options
    parser.addoption(
        "--urls-skip-libosinfo",
        action="store_true",
        default=False,
        help=(
            "For test_urls.py, "
            "Don't use libosinfo for media/tree detection, "
            "Use our internal detection logic."
        ),
    )
    parser.addoption(
        "--urls-force-libosinfo",
        action="store_true",
        default=False,
        help=(
            "For test_urls.py, Only use libosinfo for "
            "media/tree detection. This will skip "
            "some cases that are known not to work, "
            "like debian/ubuntu tree detection."
        ),
    )
    parser.addoption(
        "--urls-iso-only",
        action="store_true",
        default=False,
        help=("For test_urls.py, Only run iso tests."),
    )
    parser.addoption(
        "--urls-url-only",
        action="store_true",
        default=False,
        help=("For test_urls.py, Only run url tests"),
    )


def pytest_ignore_collect(path, config):
    uitests_requested = config.getoption("--uitests")

    # Default --uitests to --verbosity=2
    if uitests_requested:
        config.option.verbose = max(2, config.option.verbose)

    # Unless explicitly requested, ignore these tests
    if "test_dist.py" in str(path):
        return True
    if "test_urls.py" in str(path):
        return True
    if "test_inject.py" in str(path):
        return True

    uitest_file = "tests/uitests" in str(path)
    if uitest_file and not uitests_requested:
        return True
    if not uitest_file and uitests_requested:
        return True


def pytest_collection_modifyitems(config, items):
    def find_items(basename):
        return [i for i in items if os.path.basename(i.fspath) == basename]

    # Move test_cli cases to the end, because they are slow
    # Move test_checkprops to the very end, because it needs to run
    #   after everything else to give proper results
    cliitems = find_items("test_cli.py")
    chkitems = find_items("test_checkprops.py")

    for i in cliitems + chkitems:
        items.remove(i)
        items.append(i)

    if not find_items("test_urls.py"):
        # Don't setup urlfetcher mocking for test_urls.py
        # All other tests need it
        from tests import urlfetcher_mock

        urlfetcher_mock.setup_mock()

    if find_items("test_inject.py"):
        if not config.getoption("--capture") == "no":
            pytest.fail("test_inject.py requires `pytest --capture=no`")


@pytest.hookimpl(tryfirst=True, hookwrapper=True)
def pytest_runtest_makereport(item, call):
    """
    Hooks into test reporting so we can track if any tests didn't
    pass, so we can skip checkprops checks which require everything to pass.
    """
    ignore = item
    ignore = call
    outcome = yield
    testreport = outcome.get_result()
    if not testreport.passed:
        TESTCONFIG.skip_checkprops = True


def pytest_configure(config):
    TESTCONFIG.url_iso_only = config.getoption("--urls-iso-only")
    TESTCONFIG.url_only = config.getoption("--urls-url-only")
    TESTCONFIG.url_skip_libosinfo = config.getoption("--urls-skip-libosinfo")
    TESTCONFIG.url_force_libosinfo = config.getoption("--urls-force-libosinfo")
    TESTCONFIG.regenerate_output = config.getoption("--regenerate-output")

    TESTCONFIG.debug = config.getoption("--log-level") == "debug"
    tests.setup_logging()



# Copyright (C) 2011, 2013 Red Hat, Inc.
# Copyright (C) 2011 Cole Robinson <crobinso@redhat.com>
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

# This module provides a simple way to trace any activity on a specific
# python class or module. The trace output is logged using the regular
# logging infrastructure. Invoke this with virt-manager --trace-libvirt

import re
import threading
import time
import traceback
from types import FunctionType

from virtinst import log


CHECK_MAINLOOP = False


def generate_wrapper(origfunc, name):
    # This could be used as generic infrastructure, but it has hacks for
    # identifying places where libvirt hits the network from the main thread,
    # which causes UI blocking on slow network connections.

    def newfunc(*args, **kwargs):
        threadname = threading.current_thread().name
        is_main_thread = threading.current_thread().name == "MainThread"

        # These APIs don't hit the network, so we might not want to see them.
        is_non_network_libvirt_call = (
            name.endswith(".name")
            or name.endswith(".UUIDString")
            or name.endswith(".__init__")
            or name.endswith(".__del__")
            or name.endswith(".connect")
            or name.startswith("libvirtError")
        )

        if not is_non_network_libvirt_call and (is_main_thread or not CHECK_MAINLOOP):
            tb = ""
            if is_main_thread:
                tb = "\n%s" % "".join(traceback.format_stack())
            log.debug(
                "TRACE %s: thread=%s: %s %s %s%s", time.time(), threadname, name, args, kwargs, tb
            )
        return origfunc(*args, **kwargs)

    return newfunc


def wrap_func(module, funcobj):
    name = funcobj.__name__
    log.debug("wrapfunc %s %s", funcobj, name)

    newfunc = generate_wrapper(funcobj, name)
    setattr(module, name, newfunc)


def wrap_method(classobj, methodobj):
    name = methodobj.__name__
    fullname = classobj.__name__ + "." + name
    log.debug("wrapmeth %s", fullname)

    newfunc = generate_wrapper(methodobj, fullname)
    setattr(classobj, name, newfunc)


def wrap_class(classobj):
    log.debug("wrapclas %s %s", classobj, classobj.__name__)

    for name in dir(classobj):
        obj = getattr(classobj, name)
        if isinstance(obj, FunctionType):
            wrap_method(classobj, obj)


def wrap_module(module, mainloop, regex):
    global CHECK_MAINLOOP
    CHECK_MAINLOOP = mainloop
    for name in dir(module):
        if regex and not re.match(regex, name):
            continue  # pragma: no cover
        obj = getattr(module, name)
        if isinstance(obj, FunctionType):
            wrap_func(module, obj)
        if isinstance(obj, type):
            wrap_class(obj)



# Copyright (C) 2013 Red Hat, Inc.
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

import pytest

from virtinst import Guest
from virtinst import OSDB
from virtinst import xmlutil
from virtinst.install import urldetect

from tests import utils


##################
# Test osdict.py #
##################


def test_list_os():
    OSDB.list_os()


def test_recommended_resources():
    conn = utils.URIs.open_testdefault_cached()
    guest = Guest(conn)
    res = OSDB.lookup_os("generic").get_recommended_resources()
    assert res.get_recommended_ram(guest.os.arch) is None

    res = OSDB.lookup_os("fedora21").get_recommended_resources()
    assert res.get_recommended_ncpus(guest.os.arch) == 2


def test_urldetct_matching_distros():
    # pylint: disable=protected-access
    allstores = urldetect._build_distro_list(OSDB.lookup_os("generic"))

    seen_distro = []
    for store in allstores:
        for distro in store.matching_distros:
            if distro in seen_distro:
                raise xmlutil.DevError(
                    "store=%s has conflicting matching_distro=%s " % (store.PRETTY_NAME, distro)
                )
            seen_distro.append(distro)


def test_tree_url():
    f26 = OSDB.lookup_os("fedora26")
    f29 = OSDB.lookup_os("fedora29")
    winxp = OSDB.lookup_os("winxp")

    # Valid tree URL
    assert "fedoraproject.org" in f26.get_location("x86_64")

    # Most generic tree URL
    assert "Everything" in f29.get_location("x86_64")

    # Specific tree
    assert "Server" in f29.get_location("x86_64", "jeos")
    assert "Workstation" in f29.get_location("x86_64", "desktop")

    # Has tree URLs, but none for arch
    with pytest.raises(RuntimeError, match="ia64"):
        f26.get_location("ia64")

    # Has no tree URLs
    with pytest.raises(RuntimeError, match=".*URL location$"):
        winxp.get_location("x86_64")

    # Trigger an error path for code coverage
    assert OSDB.guess_os_by_tree(utils.TESTDIR) is None


def test_kernel_url():
    def _c(name):
        osobj = OSDB.lookup_os(name)
        if not osobj:
            pytest.skip("osinfo-db doesn't have '%s'" % name)
        return osobj.get_kernel_url_arg()

    assert _c("rhel7-unknown") == "inst.repo"
    assert _c("rhel6-unknown") == "method"
    assert _c("fedora-rawhide") == "inst.repo"
    assert _c("fedora20") == "inst.repo"
    assert _c("generic") is None
    assert _c("win10") is None
    assert _c("sle15") == "install"


def test_related_to():
    # pylint: disable=protected-access
    win10 = OSDB.lookup_os("win10")
    assert win10._is_related_to("winxp") is True
    assert win10._is_related_to("win10") is True
    assert win10._is_related_to("fedora26") is False


def test_drivers():
    win7 = OSDB.lookup_os("win7")
    generic = OSDB.lookup_os("generic")
    assert generic.supports_unattended_drivers("x86_64") is False
    assert win7.supports_unattended_drivers("x86_64") is True
    assert win7.supports_unattended_drivers("fakearch") is False
    assert win7.get_pre_installable_drivers_location("x86_64")



# Copyright (C) 2013, 2014 Red Hat, Inc.
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

import os

import pytest

from tests import utils

from virtinst import Capabilities
from virtinst import DomainCapabilities


DATADIR = utils.DATADIR + "/capabilities"


def _buildCaps(filename):
    path = os.path.join(DATADIR, filename)
    conn = utils.URIs.open_testdefault_cached()
    return Capabilities(conn, open(path).read())


def testCapsCPUFeaturesNewSyntax():
    filename = "test-qemu-with-kvm.xml"
    caps = _buildCaps(filename)

    assert caps.host.cpu.arch == "x86_64"
    assert caps.host.cpu.model == "core2duo"


def testCapsUtilFuncs():
    caps_with_kvm = _buildCaps("test-qemu-with-kvm.xml")
    caps_no_kvm = _buildCaps("test-qemu-no-kvm.xml")
    caps_empty = _buildCaps("test-empty.xml")

    def test_utils(caps, has_guests, is_kvm):
        assert caps.has_install_options() == has_guests
        if caps.guests:
            assert caps.guests[0].is_kvm_available() == is_kvm

    test_utils(caps_empty, False, False)
    test_utils(caps_with_kvm, True, True)
    test_utils(caps_no_kvm, True, False)

    # Small test for extra coverage
    with pytest.raises(ValueError, match=r".*virtualization type 'xen'.*"):
        caps_empty.guest_lookup(os_type="linux")
    with pytest.raises(ValueError, match=r".*not support any.*"):
        caps_empty.guest_lookup()


def testGuestCapabilities():
    filename = "kvm-x86_64.xml"
    caps = _buildCaps(filename)

    assert caps.guests[0].supports_externalSnapshot() is True


##############################
# domcapabilities.py testing #
##############################


def testDomainCapabilities():
    xml = open(DATADIR + "/test-domcaps.xml").read()
    caps = DomainCapabilities(utils.URIs.open_testdriver_cached(), xml)

    assert caps.machine == "my-machine-type"
    assert caps.arch == "x86_64"
    assert caps.domain == "kvm"
    assert caps.path == "/bin/emulatorbin"

    assert caps.os.loader.supported is True
    assert caps.os.loader.get_values() == ["/foo/bar", "/tmp/my_path"]
    assert caps.os.loader.enum_names() == ["type", "readonly"]
    assert caps.os.loader.get_enum("type").get_values() == ["rom", "pflash"]
    assert caps.os.loader.get_enum("idontexist").get_values() == []


def testDomainCapabilitiesx86():
    xml = open(DATADIR + "/kvm-x86_64-domcaps-latest.xml").read()
    caps = DomainCapabilities(utils.URIs.open_testdriver_cached(), xml)

    custom_mode = caps.cpu.get_mode("custom")
    assert bool(custom_mode)
    cpu_model = custom_mode.get_model("Opteron_G4")
    assert bool(cpu_model)
    assert cpu_model.usable

    models = caps.get_cpu_models()
    assert len(models) > 10
    assert "SandyBridge" in models

    assert caps.label_for_firmware_path(None) == "BIOS"
    assert "Custom:" in caps.label_for_firmware_path("/foobar")
    assert "UEFI" in caps.label_for_firmware_path("OVMF")

    assert caps.supports_filesystem_virtiofs()
    assert caps.supports_memorybacking_memfd()
    assert caps.supports_redirdev_usb()
    assert caps.supports_channel_spicevmc()
    assert caps.supported_panic_models() == ["isa", "hyperv", "pvpanic"]

    xml = open(DATADIR + "/kvm-x86_64-domcaps-amd-sev.xml").read()
    caps = DomainCapabilities(utils.URIs.open_testdriver_cached(), xml)
    assert caps.supports_sev_launch_security()


def testDomainCapabilitiesAArch64():
    xml = open(DATADIR + "/kvm-aarch64-domcaps.xml").read()
    caps = DomainCapabilities(utils.URIs.open_testdriver_cached(), xml)

    assert "Default" in caps.label_for_firmware_path(None)
    assert "Custom:" in caps.label_for_firmware_path("/foobar")
    assert "UEFI" in caps.label_for_firmware_path("aarch64/QEMU_EFI")

    assert caps.supports_filesystem_virtiofs()
    assert caps.supports_memorybacking_memfd()
    assert caps.supports_redirdev_usb()
    assert caps.supports_channel_spicevmc()
    assert caps.supported_panic_models() == ["pvpanic"]


def testDomainCapabilitiesPPC64le():
    xml = open(DATADIR + "/kvm-ppc64le-domcaps.xml").read()
    caps = DomainCapabilities(utils.URIs.open_testdriver_cached(), xml)

    custom_mode = caps.cpu.get_mode("custom")
    assert bool(custom_mode)

    models = caps.get_cpu_models()
    assert "POWER9" in models

    assert "Default" in caps.label_for_firmware_path(None)

    assert caps.supports_filesystem_virtiofs()
    assert caps.supports_memorybacking_memfd()
    assert caps.supports_redirdev_usb()
    assert not caps.supports_channel_spicevmc()
    assert caps.supported_panic_models() == ["pseries", "pvpanic"]


def testDomainCapabilitiesRISCV64():
    xml = open(DATADIR + "/qemu-riscv64-domcaps.xml").read()
    caps = DomainCapabilities(utils.URIs.open_testdriver_cached(), xml)

    host_mode = caps.cpu.get_mode("host-passthrough")
    assert bool(host_mode)
    assert not host_mode.supported
    max_mode = caps.cpu.get_mode("maximum")
    assert bool(max_mode)
    assert max_mode.supported
    custom_mode = caps.cpu.get_mode("custom")
    assert bool(custom_mode)
    cpu_model = custom_mode.get_model("rv64")
    assert bool(cpu_model)
    assert cpu_model.usable

    models = caps.get_cpu_models()
    assert len(models) > 5
    assert "veyron-v1" in models

    assert "Default" in caps.label_for_firmware_path(None)
    assert "Custom:" in caps.label_for_firmware_path("/foobar")
    assert "UEFI" in caps.label_for_firmware_path("RISCV_VIRT_CODE.fd")

    assert caps.supports_filesystem_virtiofs()
    assert caps.supports_memorybacking_memfd()
    assert caps.supports_redirdev_usb()
    assert caps.supports_channel_spicevmc()
    assert caps.supported_panic_models() == ["pvpanic"]


def testDomainCapabilitiesLoongArch64():
    xml = open(DATADIR + "/kvm-loongarch64-domcaps.xml").read()
    caps = DomainCapabilities(utils.URIs.open_testdriver_cached(), xml)

    host_mode = caps.cpu.get_mode("host-passthrough")
    assert bool(host_mode)
    assert host_mode.supported
    max_mode = caps.cpu.get_mode("maximum")
    assert bool(max_mode)
    assert max_mode.supported
    custom_mode = caps.cpu.get_mode("custom")
    assert bool(custom_mode)
    cpu_model = custom_mode.get_model("la132")
    assert bool(cpu_model)
    assert cpu_model.usable

    models = caps.get_cpu_models()
    assert len(models) > 2
    assert "la464" in models

    assert "Default" in caps.label_for_firmware_path(None)
    assert "Custom:" in caps.label_for_firmware_path("/foobar")
    assert "UEFI" in caps.label_for_firmware_path("loongarch64/QEMU_CODE.fd")

    assert caps.supports_filesystem_virtiofs()
    assert caps.supports_memorybacking_memfd()
    assert caps.supports_redirdev_usb()
    assert caps.supports_channel_spicevmc()
    assert caps.supported_panic_models() == ["pvpanic"]



# Copyright (C) 2006, 2012-2013 Red Hat, Inc.
# Copyright (C) 2006 Daniel P. Berrange <berrange@redhat.com>
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

from gi.repository import Gtk
from gi.repository import Gdk

from virtinst import DomainCpu
from virtinst import log
from virtinst import xmlutil

from .lib import uiutil
from .baseclass import vmmGObjectUI
from .lib.inspection import vmmInspection
from .systray import vmmSystray


class vmmPreferences(vmmGObjectUI):
    @classmethod
    def show_instance(cls, parentobj):
        try:
            if not cls._instance:
                cls._instance = vmmPreferences()
            cls._instance.show(parentobj.topwin)
        except Exception as e:  # pragma: no cover
            parentobj.err.show_err(_("Error launching preferences: %s") % str(e))

    def __init__(self):
        vmmGObjectUI.__init__(self, "preferences.ui", "vmm-preferences")
        self._cleanup_on_app_close()

        self._init_ui()

        self._orig_libguestfs_val = None

        self.refresh_view_system_tray()
        self.refresh_xmleditor()
        self.refresh_libguestfs()
        self.refresh_update_interval()
        self.refresh_console_scaling()
        self.refresh_console_resizeguest()
        self.refresh_console_autoredir()
        self.refresh_console_autoconnect()
        self.refresh_graphics_type()
        self.refresh_storage_format()
        self.refresh_cpu_default()
        self.refresh_firmware_default()
        self.refresh_cpu_poll()
        self.refresh_disk_poll()
        self.refresh_net_poll()
        self.refresh_memory_poll()
        self.refresh_grabkeys_combination()
        self.refresh_confirm_forcepoweroff()
        self.refresh_confirm_poweroff()
        self.refresh_confirm_pause()
        self.refresh_confirm_removedev()
        self.refresh_confirm_unapplied()
        self.refresh_confirm_delstorage()

        self.builder.connect_signals(
            {
                "on_vmm_preferences_delete_event": self.close,
                "on_prefs_close_clicked": self.close,
                "on_prefs_system_tray_toggled": self.change_view_system_tray,
                "on_prefs_xmleditor_toggled": self.change_xmleditor,
                "on_prefs_libguestfs_toggled": self.change_libguestfs,
                "on_prefs_stats_update_interval_changed": self.change_update_interval,
                "on_prefs_console_scaling_changed": self.change_console_scaling,
                "on_prefs_console_resizeguest_changed": self.change_console_resizeguest,
                "on_prefs_console_autoredir_changed": self.change_console_autoredir,
                "on_prefs_console_autoconnect_toggled": self.change_console_autoconnect,
                "on_prefs_graphics_type_changed": self.change_graphics_type,
                "on_prefs_storage_format_changed": self.change_storage_format,
                "on_prefs_cpu_default_changed": self.change_cpu_default,
                "on_prefs_firmware_default_changed": self.change_firmware_default,
                "on_prefs_stats_enable_cpu_toggled": self.change_cpu_poll,
                "on_prefs_stats_enable_disk_toggled": self.change_disk_poll,
                "on_prefs_stats_enable_net_toggled": self.change_net_poll,
                "on_prefs_stats_enable_memory_toggled": self.change_memory_poll,
                "on_prefs_confirm_forcepoweroff_toggled": self.change_confirm_forcepoweroff,
                "on_prefs_confirm_poweroff_toggled": self.change_confirm_poweroff,
                "on_prefs_confirm_pause_toggled": self.change_confirm_pause,
                "on_prefs_confirm_removedev_toggled": self.change_confirm_removedev,
                "on_prefs_confirm_unapplied_toggled": self.change_confirm_unapplied,
                "on_prefs_confirm_delstorage_toggled": self.change_confirm_delstorage,
                "on_prefs_btn_keys_define_clicked": self.change_grab_keys,
            }
        )

        self.widget("prefs-graphics-type").emit("changed")

        self.bind_escape_key_close()

    def close(self, ignore1=None, ignore2=None):
        log.debug("Closing preferences")
        self.topwin.hide()
        return 1

    def show(self, parent):
        log.debug("Showing preferences")
        self.topwin.set_transient_for(parent)
        self.topwin.present()

    def _cleanup(self):
        pass

    def _init_ui(self):
        combo = self.widget("prefs-console-scaling")
        # [gsettings value, string]
        model = Gtk.ListStore(int, str)
        for row in [[0, _("Never")], [1, _("Fullscreen only")], [2, _("Always")]]:
            model.append(row)
        combo.set_model(model)
        uiutil.init_combo_text_column(combo, 1)

        combo = self.widget("prefs-console-resizeguest")
        # [gsettings value, string]
        model = Gtk.ListStore(int, str)
        vals = {
            0: _("Off"),
            1: _("On"),
        }
        model.append([-1, _("System default (%s)") % vals[self.config.default_console_resizeguest]])
        for key, val in vals.items():
            model.append([key, val])
        combo.set_model(model)
        uiutil.init_combo_text_column(combo, 1)

        combo = self.widget("prefs-console-autoredir")
        # [gsettings value, string]
        model = Gtk.ListStore(bool, str)
        vals = {
            False: _("Manual redirect only"),
            True: _("Auto redirect on USB attach"),
        }
        for key, val in vals.items():
            model.append([key, val])
        combo.set_model(model)
        uiutil.init_combo_text_column(combo, 1)

        combo = self.widget("prefs-graphics-type")
        # [gsettings value, string]
        model = Gtk.ListStore(str, str)
        for row in [
            ["system", _("System default (%s)") % self.config.default_graphics_from_config],
            ["vnc", "VNC"],
            ["spice", "Spice"],
        ]:
            model.append(row)
        combo.set_model(model)
        uiutil.init_combo_text_column(combo, 1)

        combo = self.widget("prefs-storage-format")
        # [gsettings value, string]
        model = Gtk.ListStore(str, str)
        for row in [
            ["default", _("System default (%s)") % self.config.default_storage_format_from_config],
            ["raw", "Raw"],
            ["qcow2", "QCOW2"],
        ]:
            model.append(row)
        combo.set_model(model)
        uiutil.init_combo_text_column(combo, 1)

        combo = self.widget("prefs-cpu-default")
        # [gsettings value, string]
        model = Gtk.ListStore(str, str)
        for row in [
            [DomainCpu.SPECIAL_MODE_APP_DEFAULT, _("Application default")],
            [DomainCpu.SPECIAL_MODE_HV_DEFAULT, _("Hypervisor default")],
            [DomainCpu.SPECIAL_MODE_HOST_MODEL_ONLY, _("Nearest host CPU model")],
            [DomainCpu.SPECIAL_MODE_HOST_MODEL, "host-model"],
            [DomainCpu.SPECIAL_MODE_HOST_PASSTHROUGH, "host-passthrough"],
            [DomainCpu.SPECIAL_MODE_MAXIMUM, "maximum"],
        ]:
            model.append(row)
        combo.set_model(model)
        uiutil.init_combo_text_column(combo, 1)

        combo = self.widget("prefs-firmware-default")
        # [gsettings value, string]
        model = Gtk.ListStore(str, str)
        for row in [["default", _("System default")], ["uefi", "UEFI"]]:
            model.append(row)
        combo.set_model(model)
        uiutil.init_combo_text_column(combo, 1)

        if not vmmInspection.libguestfs_installed():  # pragma: no cover
            self.widget("prefs-libguestfs").set_sensitive(False)
            self.widget("prefs-libguestfs").set_tooltip_text(
                _("python libguestfs support is not installed")
            )

    #########################
    # Config Change Options #
    #########################

    def refresh_view_system_tray(self):
        errmsg = vmmSystray.systray_disabled_message()
        has_errmsg = bool(errmsg)
        val = bool(self.config.get_view_system_tray()) and not has_errmsg
        self.widget("prefs-system-tray").set_sensitive(not has_errmsg)
        self.widget("prefs-system-tray").set_active(val)
        if has_errmsg:
            self.widget("prefs-system-tray-warn-label").set_markup(
                "<small>%s</small>" % xmlutil.xml_escape(errmsg)
            )
        uiutil.set_grid_row_visible(self.widget("prefs-system-tray-warn-box"), has_errmsg)

    def refresh_xmleditor(self):
        val = self.config.get_xmleditor_enabled()
        self.widget("prefs-xmleditor").set_active(bool(val))

    def refresh_libguestfs(self):
        val = self.config.get_libguestfs_inspect_vms()
        if self._orig_libguestfs_val is None:
            self._orig_libguestfs_val = val
        self.widget("prefs-libguestfs").set_active(bool(val))

    def refresh_update_interval(self):
        self.widget("prefs-stats-update-interval").set_value(
            self.config.get_stats_update_interval()
        )

    def refresh_console_scaling(self):
        combo = self.widget("prefs-console-scaling")
        val = self.config.get_console_scaling()
        uiutil.set_list_selection(combo, val)

    def refresh_console_resizeguest(self):
        combo = self.widget("prefs-console-resizeguest")
        val = self.config.get_console_resizeguest()
        uiutil.set_list_selection(combo, val)

    def refresh_console_autoredir(self):
        combo = self.widget("prefs-console-autoredir")
        val = self.config.get_auto_usbredir()
        uiutil.set_list_selection(combo, val)

    def refresh_console_autoconnect(self):
        val = self.config.get_console_autoconnect()
        self.widget("prefs-console-autoconnect").set_active(val)

    def refresh_graphics_type(self):
        combo = self.widget("prefs-graphics-type")
        gtype = self.config.get_graphics_type(raw=True)
        uiutil.set_list_selection(combo, gtype)

    def refresh_storage_format(self):
        combo = self.widget("prefs-storage-format")
        val = self.config.get_default_storage_format(raw=True)
        uiutil.set_list_selection(combo, val)

    def refresh_cpu_default(self):
        combo = self.widget("prefs-cpu-default")
        val = self.config.get_default_cpu_setting()
        uiutil.set_list_selection(combo, val)

    def refresh_firmware_default(self):
        combo = self.widget("prefs-firmware-default")
        val = self.config.get_default_firmware_setting()
        uiutil.set_list_selection(combo, val)

    def refresh_cpu_poll(self):
        self.widget("prefs-stats-enable-cpu").set_active(self.config.get_stats_enable_cpu_poll())

    def refresh_disk_poll(self):
        self.widget("prefs-stats-enable-disk").set_active(self.config.get_stats_enable_disk_poll())

    def refresh_net_poll(self):
        self.widget("prefs-stats-enable-net").set_active(self.config.get_stats_enable_net_poll())

    def refresh_memory_poll(self):
        self.widget("prefs-stats-enable-memory").set_active(
            self.config.get_stats_enable_memory_poll()
        )

    def _process_grabkeys(self, val):
        # We convert keysyms to names
        keys = []
        for k in val.split(","):
            try:
                key = int(k)
                keys.append(Gdk.keyval_name(key))
            except Exception:  # pragma: no cover
                continue
        return "+".join(keys)

    def refresh_grabkeys_combination(self):
        val = self.config.get_keys_combination()
        keystr = self._process_grabkeys(val)
        self.widget("prefs-keys-grab-sequence").set_text(keystr or "")

    def refresh_confirm_forcepoweroff(self):
        self.widget("prefs-confirm-forcepoweroff").set_active(
            self.config.get_confirm_forcepoweroff()
        )

    def refresh_confirm_poweroff(self):
        self.widget("prefs-confirm-poweroff").set_active(self.config.get_confirm_poweroff())

    def refresh_confirm_pause(self):
        self.widget("prefs-confirm-pause").set_active(self.config.get_confirm_pause())

    def refresh_confirm_removedev(self):
        self.widget("prefs-confirm-removedev").set_active(self.config.get_confirm_removedev())

    def refresh_confirm_unapplied(self):
        self.widget("prefs-confirm-unapplied").set_active(self.config.get_confirm_unapplied())

    def refresh_confirm_delstorage(self):
        self.widget("prefs-confirm-delstorage").set_active(self.config.get_confirm_delstorage())

    def grabkeys_get_string(self, events):
        keystr = ""
        for ignore, keyval in events:
            if keystr:
                keystr += "+"
            keystr += Gdk.keyval_name(keyval)
        return keystr

    def grabkeys_dlg_press(self, src_ignore, event, label, events):
        if not [e for e in events if e[0] == event.hardware_keycode]:
            events.append((event.hardware_keycode, event.keyval))

        label.set_text(self.grabkeys_get_string(events))

    def grabkeys_dlg_release(self, src_ignore, event, label, events):
        for e in [e for e in events if e[0] == event.hardware_keycode]:
            events.remove(e)

        label.set_text(self.grabkeys_get_string(events))

    def change_grab_keys(self, src_ignore):
        dialog = Gtk.Dialog(
            _("Configure grab key combination"),
            self.topwin,
            Gtk.DialogFlags.MODAL | Gtk.DialogFlags.DESTROY_WITH_PARENT,
            (Gtk.STOCK_CANCEL, Gtk.ResponseType.REJECT, Gtk.STOCK_OK, Gtk.ResponseType.ACCEPT),
        )
        dialog.set_default_size(325, 160)
        dialog.set_border_width(6)

        infolabel = Gtk.Label(
            label=_(
                "You can now define grab keys by pressing them.\n"
                "To confirm your selection please click OK button\n"
                "while you have desired keys pressed."
            )
        )
        keylabel = Gtk.Label(label=_("Please press desired grab key combination"))

        vbox = Gtk.VBox()
        vbox.set_spacing(12)
        vbox.pack_start(infolabel, False, False, 0)
        vbox.pack_start(keylabel, False, False, 0)
        dialog.get_content_area().add(vbox)

        events = []
        dialog.connect("key-press-event", self.grabkeys_dlg_press, keylabel, events)
        dialog.connect("key-release-event", self.grabkeys_dlg_release, keylabel, events)
        dialog.show_all()
        result = dialog.run()

        if result == Gtk.ResponseType.ACCEPT:
            self.config.set_keys_combination([e[1] for e in events])

        self.refresh_grabkeys_combination()
        dialog.destroy()

    def change_view_system_tray(self, src):
        self.config.set_view_system_tray(src.get_active())

    def change_xmleditor(self, src):
        self.config.set_xmleditor_enabled(src.get_active())

    def change_libguestfs(self, src):
        val = src.get_active()
        self.config.set_libguestfs_inspect_vms(val)

        vis = val != self._orig_libguestfs_val and self.widget("prefs-libguestfs").get_sensitive()
        uiutil.set_grid_row_visible(self.widget("prefs-libguestfs-warn-box"), vis)

    def change_update_interval(self, src):
        self.config.set_stats_update_interval(src.get_value_as_int())

    def change_console_scaling(self, box):
        self.config.set_console_scaling(box.get_active())

    def change_console_resizeguest(self, box):
        val = uiutil.get_list_selection(box)
        self.config.set_console_resizeguest(val)

    def change_console_autoredir(self, box):
        val = uiutil.get_list_selection(box)
        self.config.set_auto_usbredir(val)

    def change_console_autoconnect(self, src):
        self.config.set_console_autoconnect(bool(src.get_active()))

    def change_graphics_type(self, src):
        val = uiutil.get_list_selection(src)
        self.config.set_graphics_type(val)

    def change_storage_format(self, src):
        typ = uiutil.get_list_selection(src) or "default"
        self.config.set_storage_format(typ.lower())

    def change_cpu_default(self, src):
        typ = uiutil.get_list_selection(src) or "default"
        self.config.set_default_cpu_setting(typ.lower())

    def change_firmware_default(self, src):
        typ = uiutil.get_list_selection(src) or "default"
        self.config.set_firmware_setting(typ.lower())

    def change_cpu_poll(self, src):
        self.config.set_stats_enable_cpu_poll(src.get_active())

    def change_disk_poll(self, src):
        self.config.set_stats_enable_disk_poll(src.get_active())

    def change_net_poll(self, src):
        self.config.set_stats_enable_net_poll(src.get_active())

    def change_memory_poll(self, src):
        self.config.set_stats_enable_memory_poll(src.get_active())

    def change_confirm_forcepoweroff(self, src):
        self.config.set_confirm_forcepoweroff(src.get_active())

    def change_confirm_poweroff(self, src):
        self.config.set_confirm_poweroff(src.get_active())

    def change_confirm_pause(self, src):
        self.config.set_confirm_pause(src.get_active())

    def change_confirm_removedev(self, src):
        self.config.set_confirm_removedev(src.get_active())

    def change_confirm_unapplied(self, src):
        self.config.set_confirm_unapplied(src.get_active())

    def change_confirm_delstorage(self, src):
        self.config.set_confirm_delstorage(src.get_active())



# Copyright (C) 2018 VMware, Inc.
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

from ..lib import uiutil
from ..baseclass import vmmGObjectUI


class vmmVsockDetails(vmmGObjectUI):
    __gsignals__ = {
        "changed-auto-cid": (vmmGObjectUI.RUN_FIRST, None, []),
        "changed-cid": (vmmGObjectUI.RUN_FIRST, None, []),
    }

    MIN_GUEST_CID = 3

    def __init__(self, vm, builder, topwin):
        super().__init__("vsockdetails.ui", None, builder=builder, topwin=topwin)
        self.vm = vm
        self.conn = vm.conn

        self.builder.connect_signals(
            {
                "on_vsock_auto_toggled": self._vsock_auto_toggled,
                "on_vsock_cid_changed": lambda ignore: self.emit("changed-cid"),
            }
        )

        self.top_box = self.widget("vsock-box")

    def _cleanup(self):
        self.vm = None
        self.conn = None

    ##############
    # Public API #
    ##############

    def reset_state(self):
        self.widget("vsock-auto").set_active(True)
        self.widget("vsock-cid").set_value(self.MIN_GUEST_CID)
        self.widget("vsock-cid").set_visible(False)

    def get_values(self):
        auto_cid = self.widget("vsock-auto").get_active()
        cid = uiutil.spin_get_helper(self.widget("vsock-cid"))
        return auto_cid, cid

    def set_dev(self, dev):
        self.reset_state()

        is_auto = bool(dev.auto_cid)
        cid = int(dev.cid or self.MIN_GUEST_CID)

        label = self.widget("vsock-auto").get_label().split(" (")[0]
        if is_auto and self.vm.is_active():
            label += " (%s %s)" % (_("CID"), cid)
        self.widget("vsock-auto").set_label(label)

        self.widget("vsock-auto").set_active(is_auto)
        self.widget("vsock-cid").set_value(cid)
        self.widget("vsock-cid").set_visible(not is_auto)

    #############
    # Listeners #
    #############

    def _vsock_auto_toggled(self, ignore):
        is_auto = self.widget("vsock-auto").get_active()
        self.widget("vsock-cid").set_visible(not is_auto)
        self.emit("changed-auto-cid")



# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

import re

from gi.repository import Gdk

import dogtail.tree
import pyatspi

from virtinst import log
from . import utils


class _FuzzyPredicate(dogtail.predicate.Predicate):
    """
    Object dogtail/pyatspi want for node searching.
    """

    def __init__(
        self, name=None, roleName=None, labeller_text=None, focusable=False, onscreen=False
    ):
        """
        :param name: Match node.name or node.labeller.text if
            labeller_text not specified
        :param roleName: Match node.roleName
        :param labeller_text: Match node.labeller.text
        :param focusable: Ensure node is focusable
        """
        self._name = name
        self._roleName = roleName
        self._labeller_text = labeller_text
        self._focusable = focusable
        self._onscreen = onscreen

        self._name_pattern = None
        self._role_pattern = None
        self._labeller_pattern = None
        if self._name:
            self._name_pattern = re.compile(self._name, re.DOTALL)
        if self._roleName:
            self._role_pattern = re.compile(self._roleName, re.DOTALL)
        if self._labeller_text:
            self._labeller_pattern = re.compile(self._labeller_text, re.DOTALL)

    def makeScriptMethodCall(self, isRecursive):
        ignore = isRecursive
        return

    def makeScriptVariableName(self):
        return

    def describeSearchResult(self, node=None):
        if not node:
            return ""
        return node.node_string()

    def satisfiedByNode(self, node):
        """
        The actual search routine
        """
        try:
            if self._roleName and not self._role_pattern.match(node.roleName):
                return

            labeller = ""
            if node.labeller:
                labeller = node.labeller.text

            if (
                self._name
                and not self._name_pattern.match(node.name)
                and not self._name_pattern.match(labeller)
            ):
                return
            if self._labeller_text and not self._labeller_pattern.match(labeller):
                return
            if self._focusable and not (
                node.focusable
                and node.onscreen
                and node.sensitive
                and node.roleName not in ["page tab list", "radio button"]
            ):
                return False
            return True
        except Exception as e:
            log.debug(
                "got predicate exception name=%s role=%s labeller=%s: %s",
                self._name,
                self._roleName,
                self._labeller_text,
                e,
            )


def _debug_decorator(fn):
    def _cb(self, *args, **kwargs):
        try:
            return fn(self, *args, **kwargs)
        except Exception:
            print("node=%s\nstates=%s" % (self, self.print_states()))
            raise

    return _cb


class _VMMDogtailNode(dogtail.tree.Node):
    """
    Our extensions to the dogtail node wrapper class.
    """

    # The class hackery means pylint can't figure this class out
    # pylint: disable=no-member

    @property
    def active(self):
        """
        If the window is the raised and active window or not
        """
        return self.getState().contains(pyatspi.STATE_ACTIVE)

    @property
    def state_selected(self):
        return self.getState().contains(pyatspi.STATE_SELECTED)

    @property
    def onscreen(self):
        # We need to check that full widget is on screen because we use this
        # function to check whether we can click a widget. We may click
        # anywhere within the widget and clicks outside the screen bounds are
        # silently ignored.
        if self.roleName in ["frame"]:
            return True
        screen = Gdk.Screen.get_default()
        return (
            self.position[0] >= 0
            and self.position[0] + self.size[0] < screen.get_width()
            and self.position[1] >= 0
            and self.position[1] + self.size[1] < screen.get_height()
        )

    @_debug_decorator
    def check_onscreen(self):
        """
        Check in a loop that the widget is onscreen
        """
        utils.check(lambda: self.onscreen)

    @_debug_decorator
    def check_not_onscreen(self):
        """
        Check in a loop that the widget is not onscreen
        """
        utils.check(lambda: not self.onscreen)

    @_debug_decorator
    def check_focused(self):
        """
        Check in a loop that the widget is focused
        """
        utils.check(lambda: self.focused)

    @_debug_decorator
    def check_sensitive(self):
        """
        Check whether interactive widgets are sensitive or not
        """
        valid_types = [
            "button",
            "push button",
            "toggle button",
            "check button",
            "combo box",
            "menu item",
            "text",
            "menu",
        ]
        if self.roleName not in valid_types:
            return True
        utils.check(lambda: self.sensitive)

    def click_secondary_icon(self):
        """
        Helper for clicking the secondary icon of a text entry
        """
        self.check_onscreen()
        self.check_sensitive()
        button = 1
        clickX = self.position[0] + self.size[0] - 10
        clickY = self.position[1] + (self.size[1] / 2)
        dogtail.rawinput.click(clickX, clickY, button)

    def click_combo_entry(self):
        """
        Helper for clicking the arrow of a combo entry, to expose the menu.
        Clicks middle of Y axis, but 1/10th of the height from the right side.
        Using a small, hardcoded offset may not work on some themes (e.g. when
        running virt-manager on KDE)
        """
        self.check_onscreen()
        self.check_sensitive()
        button = 1
        clickX = self.position[0] + self.size[0] - self.size[1] / 4
        clickY = self.position[1] + self.size[1] / 2
        dogtail.rawinput.click(clickX, clickY, button)

    def click_expander(self):
        """
        Helper for clicking expander, hitting the text part to actually
        open it. Basically clicks top left corner with some indent
        """
        self.check_onscreen()
        self.check_sensitive()
        button = 1
        clickX = self.position[0] + 10
        clickY = self.position[1] + 5
        dogtail.rawinput.click(clickX, clickY, button)

    def title_coordinates(self):
        """
        Return clickable coordinates of a window's titlebar
        """
        x = self.position[0] + (self.size[0] / 2)
        y = self.position[1] + 10
        return x, y

    def click_title(self):
        """
        Helper to click a window title bar, hitting the horizontal
        center of the bar
        """
        if self.roleName not in ["frame", "alert"]:
            raise RuntimeError("Can't use click_title() on type=%s" % self.roleName)
        button = 1
        clickX, clickY = self.title_coordinates()
        dogtail.rawinput.click(clickX, clickY, button)

    def is_menuitem(self):
        submenu = self.roleName == "menu" and (
            not self.accessible_parent or self.accessible_parent.roleName == "menu"
        )
        return submenu or self.roleName == "menu item"

    def click(self, *args, **kwargs):
        """
        click wrapper, check some states first to reduce flakiness
        """
        # pylint: disable=arguments-differ,signature-differs
        self.check_onscreen()
        self.check_sensitive()
        if self.is_menuitem():
            self.point()
        super().click(*args, **kwargs)

    def point(self, *args, **kwargs):
        # pylint: disable=signature-differs
        super().point(*args, **kwargs)

        if self.is_menuitem():
            # Make sure item is selected before we return to caller
            utils.check(lambda: self.state_selected)

    def set_text(self, text):
        self.check_onscreen()
        self.check_sensitive()
        assert hasattr(self, "text")
        self.text = text

    def get_text_override(self):
        self.check_onscreen()
        self.check_sensitive()
        assert hasattr(self, "text")
        return self.text

    def bring_on_screen(self, key_name="Down", max_tries=100):
        """
        Attempts to bring the item to screen by repeatedly clicking the given
        key. Raises exception if max_tries attempts are exceeded.
        """
        cur_try = 0
        while not self.onscreen:
            dogtail.rawinput.pressKey(key_name)
            cur_try += 1
            if cur_try > max_tries:
                raise RuntimeError("Could not bring widget on screen")
        return self

    def window_maximize(self):
        assert self.roleName in ["frame", "dialog"]
        self.grab_focus()
        s1 = self.size
        self.keyCombo("<alt>F10")
        utils.check(lambda: self.size != s1)
        self.grab_focus()

    def window_close(self):
        assert self.roleName in ["frame", "alert", "dialog", "file chooser"]
        self.grab_focus()
        self.keyCombo("<alt>F4")
        utils.check(lambda: not self.showing)

    def window_find_focusable_child(self):
        return self.find(None, focusable=True)

    def grab_focus(self):
        if self.roleName in ["frame", "alert", "dialog", "file chooser"]:
            child = self.window_find_focusable_child()
            child.grab_focus()
            utils.check(lambda: self.active)
            return

        self.check_onscreen()
        assert self.focusable
        self.grabFocus()
        self.check_focused()

    #########################
    # Widget search helpers #
    #########################

    def find(
        self,
        name,
        roleName=None,
        labeller_text=None,
        check_active=True,
        recursive=True,
        focusable=False,
    ):
        """
        Search root for any widget that contains the passed name/role regex
        strings.
        """
        # Somewhere in fedora 41 stack, `push button` became just `button`
        if roleName in ["push button", ".*push button.*"]:
            roleName = "(button|push button)"
        pred = _FuzzyPredicate(name, roleName, labeller_text, focusable)

        try:
            ret = self.findChild(pred, recursive=recursive)
        except dogtail.tree.SearchError:
            raise dogtail.tree.SearchError(
                "Didn't find widget with name='%s' "
                "roleName='%s' labeller_text='%s'" % (name, roleName, labeller_text)
            ) from None

        # Wait for independent windows to become active in the window manager
        # before we return them. This ensures the window is actually onscreen
        # so it sidesteps a lot of race conditions
        if ret.roleName in ["frame", "dialog", "alert"] and check_active:
            utils.check(lambda: ret.active)
        return ret

    def find_fuzzy(self, name, roleName=None, labeller_text=None):
        """
        Search root for any widget that contains the passed name/role strings.
        """
        name_pattern = None
        role_pattern = None
        labeller_pattern = None
        if name:
            name_pattern = ".*%s.*" % name
        if roleName:
            role_pattern = ".*%s.*" % roleName
        if labeller_text:
            labeller_pattern = ".*%s.*" % labeller_text

        return self.find(name_pattern, role_pattern, labeller_pattern)

    ##########################
    # Higher level behaviors #
    ##########################

    def combo_select(self, combolabel, itemlabel):
        """
        Lookup the combo, click it, select the menu item
        """
        combo = self.find(combolabel, "combo box")
        combo.click_combo_entry()
        combo.find(itemlabel, "menu item").click()

    def combo_check_default(self, combolabel, itemlabel):
        """
        Lookup the combo and verify the menu item is selected
        """
        combo = self.find(combolabel, "combo box")
        combo.click_combo_entry()
        item = combo.find(itemlabel, "menu item")
        utils.check(lambda: item.selected)
        dogtail.rawinput.pressKey("Escape")

    #####################
    # Debugging helpers #
    #####################

    def node_string(self):
        msg = "name='%s' roleName='%s'" % (self.name, self.roleName)
        if self.labeller:
            msg += " labeller.text='%s'" % self.labeller.text
        return msg

    def fmt_nodes(self):
        strs = []

        def _walk(node):
            try:
                strs.append(node.node_string())
            except Exception as e:
                strs.append("got exception: %s" % e)

        self.findChildren(_walk, isLambda=True)
        return "\n".join(strs)

    def print_nodes(self):
        """
        Helper to print the entire node tree for the passed root. Useful
        if to figure out the roleName for the object you are looking for
        """
        print(self.fmt_nodes())

    def print_states(self):
        print([s.value_nick for s in self.getState().get_states()])


# This is the same hack dogtail uses to extend the Accessible class.
_bases = list(pyatspi.Accessibility.Accessible.__bases__)
_bases.insert(_bases.index(dogtail.tree.Node), _VMMDogtailNode)
_bases.remove(dogtail.tree.Node)
pyatspi.Accessibility.Accessible.__bases__ = tuple(_bases)



# Copyright (C) 2013 Red Hat, Inc.
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

import glob
import subprocess
import xml.etree.ElementTree as ET


def test_validate_po_files():
    """
    Validate that po translations don't mess up python format strings,
    which has broken the app in the past:
    https://bugzilla.redhat.com/show_bug.cgi?id=1350185
    https://bugzilla.redhat.com/show_bug.cgi?id=1433800
    """
    failures = []
    for pofile in glob.glob("po/*.po"):
        proc = subprocess.Popen(
            ["msgfmt", "--output-file=/dev/null", "--check", pofile],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        ignore, stderr = proc.communicate()
        if proc.wait():
            failures.append("%s: %s" % (pofile, stderr))

    if not failures:
        return

    msg = "The following po files have errors:\n"
    msg += "\n".join(failures)
    raise AssertionError(msg)


def test_validate_pot_strings():
    """
    Validate that xgettext via don't print any warnings.
    """
    potfile = "po/virt-manager.pot"
    origpot = open(potfile).read()
    try:
        subprocess.run(["meson", "setup", "build"], check=True)
        out = subprocess.check_output(
            ["meson", "compile", "-C", "build", "virt-manager-pot"], stderr=subprocess.STDOUT
        )
        warnings = [line for line in out.decode("utf-8").splitlines() if "warning:" in line]
        warnings = [warning for warning in warnings if "a fallback ITS rule file" not in warning]
        if warnings:
            raise AssertionError("xgettext has warnings:\n\n%s" % "\n".join(warnings))
    finally:
        open(potfile, "w").write(origpot)


def test_ui_minimum_version():
    """
    Ensure all glade XML files don't _require_ UI bits later than
    our minimum supported version
    """
    # Minimum dep is 3.22 to fix popups on some wayland window managers.
    # 3.22 is from Sep 2016, so coupled with python3 deps this seems fine
    # to enforce
    minimum_version_major = 3
    minimum_version_minor = 22
    minimum_version_str = "%s.%s" % (minimum_version_major, minimum_version_minor)

    failures = []
    for filename in glob.glob("ui/*.ui"):
        required_version = None
        for line in open(filename).readlines():
            # This is much faster than XML parsing the whole file
            if not line.strip().startswith("<requires "):
                continue

            req = ET.fromstring(line)
            if req.tag != "requires" or req.attrib.get("lib") != "gtk+":
                continue
            required_version = req.attrib["version"]

        if required_version is None:
            raise AssertionError("ui file=%s doesn't have a <requires> tag for gtk+")

        if (
            int(required_version.split(".")[0]) != minimum_version_major
            or int(required_version.split(".")[1]) != minimum_version_minor
        ):
            failures.append((filename, required_version))

    if not failures:
        return

    err = "The following files should require version of gtk-%s:\n" % minimum_version_str
    err += "\n".join([("%s version=%s" % tup) for tup in failures])
    raise AssertionError(err)


def test_ui_translatable_atknames():
    """
    We only use accessible names for uitests, they shouldn't be
    marked as translatable
    """
    failures = []
    atkstr = "AtkObject::accessible-name"
    for filename in glob.glob("ui/*.ui"):
        for line in open(filename).readlines():
            if atkstr not in line:
                continue
            if "translatable=" in line:
                failures.append(filename)
                break

    if not failures:
        return
    err = "Some files incorrectly have translatable ATK names.\n"
    err += "Run this command to fix:\n\n"
    err += """sed -i -e 's/%s" translatable="yes"/%s"/g' """ % (atkstr, atkstr)
    err += " ".join(failures)
    raise AssertionError(err)


def test_appstream_validate():
    subprocess.check_call(["appstream-util", "validate", "data/virt-manager.appdata.xml.in"])



# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

import time

import dogtail.config


def check(func, timeout=2):
    """
    Run the passed func in a loop every .1 seconds until timeout is hit or
    the func returns True.
    """
    start_time = time.time()
    interval = 0.1
    while True:
        if func() is True:
            return
        if (time.time() - start_time) > timeout:
            raise RuntimeError("Loop condition wasn't met")
        time.sleep(interval)


class dogtail_timeout:
    """
    Context helper to run a specific check with custom timeout, in seconds
    """

    def __init__(self, timeout):
        backoff = dogtail.config.config.searchBackoffDuration
        self._tmpval = int(timeout / backoff)
        self._origval = dogtail.config.config.searchCutoffCount

    def _set(self, val):
        dogtail.config.config.searchCutoffCount = val

    def __enter__(self):
        self._set(self._tmpval)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._set(self._origval)


def walkUIList(app, win, lst, error_cb, reverse=False):
    """
    Toggle down through a UI list like addhardware, net/storage/iface
    lists, and ensure an error isn't raised.
    """
    # Walk the lst UI and find all labelled table cells, these are
    # the actual list entries
    all_cells = lst.findChildren(lambda w: w.roleName == "table cell")
    if reverse:
        all_cells.reverse()
    all_cells[0].point()
    all_cells[0].click()
    cells_per_selection = len([c for c in all_cells if c.focused])

    idx = 0
    while idx < len(all_cells):
        cell = all_cells[idx]
        if not cell.state_selected:
            # Could be a separator table cell. Try to figure it out
            if not any([c.name for c in all_cells[idx : (idx + cells_per_selection)]]):
                idx += cells_per_selection
                continue

        check(lambda: cell.state_selected)
        app.rawinput.pressKey(reverse and "Up" or "Down")

        if not win.active:
            # Should mean an error dialog popped up
            app.root.find("Error", "alert")
            raise AssertionError("Error dialog raised?")
        if error_cb():
            raise AssertionError("Error found on a page")

        idx += cells_per_selection
        if idx >= len(all_cells):
            # Last cell, selection shouldn't have changed
            check(lambda: cell.state_selected)
        else:
            check(lambda: not cell.state_selected)


def test_xmleditor_interactions(app, win, finish):
    """
    Helper to test some common XML editor interactions
    """
    # Click the tab, make a bogus XML edit
    win.find("XML", "page tab").click()
    xmleditor = win.find("XML editor")
    xmleditor.set_text(xmleditor.text.replace("<", "<FOO", 1))

    # Trying to click away should warn that there's unapplied changes
    win.find("Details", "page tab").click()
    # Select 'No', meaning don't abandon changes
    app.click_alert_button("changes will be lost", "No")
    check(lambda: xmleditor.showing)

    # Click the finish button, but our bogus change should trigger error
    finish.click()
    app.click_alert_button("(xmlParseDoc|tag.mismatch)", "Close")

    # Try unapplied changes again, this time abandon our changes
    win.find("Details", "page tab").click()
    app.click_alert_button("changes will be lost", "Yes")
    check(lambda: not xmleditor.showing)


def get_xmleditor_xml(_app, win):
    win.find("XML", "page tab").click()
    xmleditor = win.find("XML editor")
    xml = xmleditor.get_text_override()
    win.find("Details", "page tab").click()
    check(lambda: not xmleditor.showing)
    return xml



# Copyright (C) 2006-2007, 2013, 2014 Red Hat, Inc.
# Copyright (C) 2006 Hugh O. Brock <hbrock@redhat.com>
# Copyright (C) 2014 SUSE LINUX Products GmbH, Nuernberg, Germany.
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

from gi.repository import Gtk

from virtinst import DeviceFilesystem
from virtinst import xmlutil

from ..lib import uiutil
from ..baseclass import vmmGObjectUI
from ..storagebrowse import vmmStorageBrowser


_EDIT_FS_ENUM = range(1, 8)
(
    _EDIT_FS_TYPE,
    _EDIT_FS_DRIVER,
    _EDIT_FS_SOURCE,
    _EDIT_FS_RAM_SOURCE,
    _EDIT_FS_READONLY,
    _EDIT_FS_TARGET,
    _EDIT_FS_FORMAT,
) = _EDIT_FS_ENUM


class vmmFSDetails(vmmGObjectUI):
    __gsignals__ = {"changed": (vmmGObjectUI.RUN_FIRST, None, [])}

    def __init__(self, vm, builder, topwin):
        vmmGObjectUI.__init__(self, "fsdetails.ui", None, builder=builder, topwin=topwin)

        self.vm = vm
        self.conn = vm.conn

        self._storage_browser = None
        self._active_edits = []

        def _e(edittype):
            def signal_cb(*args):
                self._change_cb(edittype)

            return signal_cb

        self.builder.connect_signals(
            {
                "on_fs_source_browse_clicked": self._browse_fs_source_cb,
                "on_fs_type_combo_changed": _e(_EDIT_FS_TYPE),
                "on_fs_driver_combo_changed": _e(_EDIT_FS_DRIVER),
                "on_fs_readonly_toggled": _e(_EDIT_FS_READONLY),
                "on_fs_format_combo_changed": _e(_EDIT_FS_FORMAT),
                "on_fs_source_changed": _e(_EDIT_FS_SOURCE),
                "on_fs_ram_source_changed": _e(_EDIT_FS_RAM_SOURCE),
                "on_fs_target_changed": _e(_EDIT_FS_TARGET),
            }
        )

        self._init_ui()
        self.top_box = self.widget("vmm-fs-details")

    def _cleanup(self):
        self.vm = None
        self.conn = None

        if self._storage_browser:
            self._storage_browser.cleanup()
            self._storage_browser = None

    ##########################
    # Initialization methods #
    ##########################

    def _init_ui(self):
        def simple_store_set(comboname, values):
            combo = self.widget(comboname)
            # [XML value, label]
            model = Gtk.ListStore(str, str)
            combo.set_model(model)
            uiutil.init_combo_text_column(combo, 1)

            for xmlval in values:
                label = xmlval
                if xmlval is None:
                    label = _("Hypervisor default")
                model.append([xmlval, label])

        # Filesystem widgets
        if self.conn.is_container_only():
            simple_store_set(
                "fs-type-combo",
                [
                    DeviceFilesystem.TYPE_MOUNT,
                    DeviceFilesystem.TYPE_FILE,
                    DeviceFilesystem.TYPE_BLOCK,
                    DeviceFilesystem.TYPE_RAM,
                ],
            )
        else:
            simple_store_set("fs-type-combo", [DeviceFilesystem.TYPE_MOUNT])

        if self.conn.is_container_only():
            simple_store_set(
                "fs-driver-combo", [DeviceFilesystem.DRIVER_LOOP, DeviceFilesystem.DRIVER_NBD, None]
            )
        else:
            domcaps = self.vm.get_domain_capabilities()
            rows = []
            if domcaps.supports_filesystem_virtiofs():
                rows.append(["virtiofs", "virtiofs"])
            rows.append([None, "virtio-9p"])
            uiutil.build_simple_combo(self.widget("fs-driver-combo"), rows, sort=False)

        simple_store_set("fs-format-combo", ["raw", "qcow2"])
        self.widget("fs-readonly").set_visible(
            self.conn.is_qemu() or self.conn.is_test() or self.conn.is_lxc()
        )

    ##############
    # UI syncing #
    ##############

    def _sync_ui(self):
        fstype = uiutil.get_list_selection(self.widget("fs-type-combo"))
        fsdriver = uiutil.get_list_selection(self.widget("fs-driver-combo"))
        is_qemu = self.conn.is_qemu() or self.conn.is_test()

        show_ram_source = fstype == DeviceFilesystem.TYPE_RAM
        uiutil.set_grid_row_visible(self.widget("fs-ram-source-box"), show_ram_source)
        uiutil.set_grid_row_visible(self.widget("fs-source-box"), not show_ram_source)

        show_format = bool(fsdriver == DeviceFilesystem.DRIVER_NBD)
        uiutil.set_grid_row_visible(self.widget("fs-format-combo"), show_format)

        show_driver_combo = is_qemu or fstype == DeviceFilesystem.TYPE_FILE

        if fstype == DeviceFilesystem.TYPE_TEMPLATE:
            source_text = _("Te_mplate:")
        else:
            source_text = _("_Source path:")

        self.widget("fs-source-title").set_text(source_text)
        self.widget("fs-source-title").set_use_underline(True)
        uiutil.set_grid_row_visible(self.widget("fs-type-combo"), not is_qemu)
        uiutil.set_grid_row_visible(self.widget("fs-driver-combo"), show_driver_combo)

        need_shared_mem = fsdriver == "virtiofs"
        have_shared_mem, _shared_mem_err = self.vm.has_shared_mem()
        show_shared_mem_warn = need_shared_mem and not have_shared_mem
        uiutil.set_grid_row_visible(self.widget("fs-driver-warn-box"), show_shared_mem_warn)
        if show_shared_mem_warn:
            label = _("You may need to 'Enable shared memory' on the 'Memory' screen.")
            self.widget("fs-driver-warn").set_markup(
                "<small>%s</small>" % xmlutil.xml_escape(label)
            )

    ##############
    # Public API #
    ##############

    def reset_state(self):
        self.widget("fs-type-combo").set_active(0)
        self.widget("fs-driver-combo").set_active(0)
        self.widget("fs-format-combo").set_active(0)
        self.widget("fs-source").set_text("")
        self.widget("fs-target").set_text("")
        self.widget("fs-readonly").set_active(False)
        self._sync_ui()
        self._active_edits = []

    def set_dev(self, dev):
        self.reset_state()

        uiutil.set_list_selection(self.widget("fs-type-combo"), dev.type)
        uiutil.set_list_selection(self.widget("fs-driver-combo"), dev.driver_type)
        uiutil.set_list_selection(self.widget("fs-format-combo"), dev.driver_format)

        if dev.type != DeviceFilesystem.TYPE_RAM:
            self.widget("fs-source").set_text(dev.source or "")
        else:
            self.widget("fs-ram-source-spin").set_value(int(dev.source) // 1024)
        self.widget("fs-target").set_text(dev.target or "")
        self.widget("fs-readonly").set_active(dev.readonly)

        self._active_edits = []

    ###################
    # Device building #
    ###################

    def _set_values(self, dev):
        fstype = uiutil.get_list_selection(self.widget("fs-type-combo"))
        usage = uiutil.spin_get_helper(self.widget("fs-ram-source-spin"))

        source = self.widget("fs-source").get_text()
        target = self.widget("fs-target").get_text()
        readonly = self.widget("fs-readonly").get_active()

        fsformat = uiutil.get_list_selection(self.widget("fs-format-combo"))
        if not self.widget("fs-format-combo").get_visible():
            fsformat = None

        driver = uiutil.get_list_selection(self.widget("fs-driver-combo"))
        if not self.widget("fs-driver-combo").get_visible():
            driver = None

        if _EDIT_FS_TYPE in self._active_edits:
            dev.type = fstype
        if _EDIT_FS_RAM_SOURCE in self._active_edits or _EDIT_FS_SOURCE in self._active_edits:
            if fstype == DeviceFilesystem.TYPE_RAM:
                dev.source = usage
                dev.source_units = "MiB"
            else:
                dev.source = source
        if _EDIT_FS_TARGET in self._active_edits:
            dev.target = target
        if _EDIT_FS_READONLY in self._active_edits:
            dev.readonly = readonly
        if _EDIT_FS_DRIVER in self._active_edits:
            origdriver = dev.driver_type
            dev.driver_type = driver
            if origdriver == "virtiofs" or driver == "virtiofs":
                # Need to reset the accessmode for virtiofs
                dev.accessmode = dev.default_accessmode()
        if _EDIT_FS_FORMAT in self._active_edits:
            dev.driver_format = fsformat

    def build_device(self):
        self._active_edits = _EDIT_FS_ENUM[:]

        conn = self.conn.get_backend()
        dev = DeviceFilesystem(conn)
        self._set_values(dev)

        dev.validate_target(dev.target)
        dev.validate()
        return dev

    def update_device(self, dev):
        newdev = DeviceFilesystem(dev.conn, parsexml=dev.get_xml())
        self._set_values(newdev)
        return newdev

    ####################
    # Internal helpers #
    ####################

    def _browse_file(self, textent, isdir=False):
        def set_storage_cb(src, path):
            if path:
                textent.set_text(path)

        reason = isdir and vmmStorageBrowser.REASON_FS or vmmStorageBrowser.REASON_IMAGE

        if self._storage_browser is None:
            self._storage_browser = vmmStorageBrowser(self.conn)

        self._storage_browser.set_finish_cb(set_storage_cb)
        self._storage_browser.set_browse_reason(reason)
        self._storage_browser.show(self.topwin.get_ancestor(Gtk.Window))

    #############
    # Listeners #
    #############

    def _change_cb(self, edittype):
        self._sync_ui()
        if edittype not in self._active_edits:
            self._active_edits.append(edittype)
        self.emit("changed")

    def _browse_fs_source_cb(self, src):
        self._browse_file(self.widget("fs-source"), isdir=True)



#
# List of OS Specific data
#
# Copyright 2006-2008, 2013-2014 Red Hat, Inc.
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

import datetime
import os
import re

from gi.repository import Libosinfo

from . import xmlutil
from .logger import log


def _media_create_from_location(location):
    if not hasattr(Libosinfo.Media, "create_from_location_with_flags"):
        return Libosinfo.Media.create_from_location(location, None)  # pragma: no cover

    # We prefer this API, because by default it will not
    # reject non-bootable media, like debian s390x
    # pylint: disable=no-member
    return Libosinfo.Media.create_from_location_with_flags(location, None, 0)


class _OsinfoIter:
    """
    Helper to turn osinfo style get_length/get_nth lists into python
    iterables
    """

    def __init__(self, listobj):
        self.current = 0
        self.listobj = listobj
        self.high = -1
        if self.listobj:
            self.high = self.listobj.get_length() - 1

    def __iter__(self):
        return self

    def __next__(self):
        if self.current > self.high:
            raise StopIteration
        ret = self.listobj.get_nth(self.current)
        self.current += 1
        return ret


class _OSDB:
    """
    Entry point for the public API
    """

    def __init__(self):
        self.__os_loader = None
        self.__os_generic = None

    #################
    # Internal APIs #
    #################

    @property
    def _os_generic(self):
        if not self.__os_generic:
            # Add our custom generic variant
            o = Libosinfo.Os()
            o.set_param("short-id", "generic")
            o.set_param("name", _("Generic or unknown OS. Usage is not recommended."))
            self.__os_generic = _OsVariant(o)
        return self.__os_generic

    @property
    def _os_loader(self):
        if not self.__os_loader:
            loader = Libosinfo.Loader()
            loader.process_default_path()

            self.__os_loader = loader
        return self.__os_loader

    @property
    def _os_db(self):
        return self._os_loader.get_db()

    ###############
    # Public APIs #
    ###############

    def lookup_os_by_full_id(self, full_id, raise_error=False):
        osobj = self._os_db.get_os(full_id)
        if osobj is None:
            if raise_error:
                raise ValueError(_("Unknown libosinfo ID '%s'") % full_id)
            return None
        return _OsVariant(osobj)

    def lookup_os(self, key, raise_error=False):
        if key == self._os_generic.name:
            return self._os_generic

        flt = Libosinfo.Filter()
        flt.add_constraint(Libosinfo.PRODUCT_PROP_SHORT_ID, key)
        oslist = self._os_db.get_os_list().new_filtered(flt).get_elements()
        if len(oslist) == 0:
            if raise_error:
                raise ValueError(
                    _("Unknown OS name '%s'. See `--osinfo list` for valid values.") % key
                )
            return None
        return _OsVariant(oslist[0])

    def guess_os_by_iso(self, location):
        try:
            media = _media_create_from_location(location)
        except Exception as e:
            log.debug("Error creating libosinfo media object: %s", str(e))
            return None

        if not self._os_db.identify_media(media):
            return None
        return media.get_os().get_short_id(), _OsMedia(media)

    def guess_os_by_tree(self, location):
        if location.startswith("/"):
            location = "file://" + location

        if xmlutil.in_testsuite() and not location.startswith("file:"):
            # We have mock network tests, but we don't want to pass the
            # fake URL to libosinfo because it slows down the testcase
            return None

        try:
            tree = Libosinfo.Tree.create_from_location(location, None)
        except Exception as e:
            log.debug("Error creating libosinfo tree object for location=%s : %s", location, str(e))
            return None

        if hasattr(self._os_db, "identify_tree"):
            # osinfo_db_identify_tree is part of libosinfo 1.6.0
            if not self._os_db.identify_tree(tree):
                return None  # pragma: no cover
            return tree.get_os().get_short_id(), _OsTree(tree)
        else:  # pragma: no cover
            osobj, treeobj = self._os_db.guess_os_from_tree(tree)
            if not osobj:
                return None  # pragma: no cover
            return osobj.get_short_id(), _OsTree(treeobj)

    def list_os(self, sortkey="name"):
        """
        List all OSes in the DB, sorting by the passes _OsVariant attribute
        """
        oslist = [_OsVariant(osent) for osent in self._os_db.get_os_list().get_elements()]
        oslist.append(self._os_generic)

        # human/natural sort, but with reverse sorted numbers
        def to_int(text):
            return (int(text) * -1) if text.isdigit() else text.lower()

        def alphanum_key(obj):
            val = getattr(obj, sortkey)
            return [to_int(c) for c in re.split("([0-9]+)", val)]

        return list(sorted(oslist, key=alphanum_key))


OSDB = _OSDB()


#####################
# OsResources class #
#####################


class _OsResources:
    def __init__(self, minimum, recommended):
        self._minimum = self._convert_to_dict(minimum)
        self._recommended = self._convert_to_dict(recommended)

    def _convert_to_dict(self, resources):
        """
        Convert an OsResources object to a dictionary for easier
        lookups. Layout is: {arch: {strkey: value}}
        """
        ret = {}
        for r in _OsinfoIter(resources):
            vals = {}
            vals["ram"] = r.get_ram()
            vals["n-cpus"] = r.get_n_cpus()
            vals["storage"] = r.get_storage()
            ret[r.get_architecture()] = vals
        return ret

    def _get_key(self, resources, key, arch):
        for checkarch in [arch, "all"]:
            val = resources.get(checkarch, {}).get(key, -1)
            if val != -1:
                return val

    def _get_minimum_key(self, key, arch):
        val = self._get_key(self._minimum, key, arch)
        if val and val > 0:
            return val

    def _get_recommended_key(self, key, arch):
        val = self._get_key(self._recommended, key, arch)
        if val and val > 0:
            return val
        # If we are looking for a recommended value, but the OS
        # DB only has minimum resources tracked, double the minimum
        # value as an approximation at a 'recommended' value
        val = self._get_minimum_key(key, arch)
        if val:
            log.debug("No recommended value found for key='%s', using minimum=%s * 2", key, val)
            return val * 2
        return None

    def get_minimum_ram(self, arch):
        return self._get_minimum_key("ram", arch)

    def get_recommended_ram(self, arch):
        return self._get_recommended_key("ram", arch)

    def get_recommended_ncpus(self, arch):
        return self._get_recommended_key("n-cpus", arch)

    def get_recommended_storage(self, arch):
        return self._get_recommended_key("storage", arch)


#####################
# OsVariant classes #
#####################


class _OsVariant:
    def __init__(self, o):
        self._os = o

        self._short_ids = [self._os.get_short_id()]
        if hasattr(self._os, "get_short_id_list"):
            self._short_ids = self._os.get_short_id_list()
        self.name = self._short_ids[0]
        self.all_names = list(sorted(set(self._short_ids)))

        self._family = self._os.get_family()
        self.full_id = self._os.get_id()
        self.label = self._os.get_name()
        self.codename = self._os.get_codename() or ""
        self.distro = self._os.get_distro() or ""
        self.version = self._os.get_version()

        self.eol = self._get_eol()

    def __repr__(self):
        return "<%s name=%s>" % (self.__class__.__name__, self.name)

    ########################
    # Internal helper APIs #
    ########################

    def _is_related_to(
        self,
        related_os_list,
        osobj=None,
        check_derives=True,
        check_upgrades=True,
        check_clones=True,
    ):
        osobj = osobj or self._os
        if osobj.get_short_id() in related_os_list:
            return True

        check_list = []

        def _extend(newl):
            for obj in newl:
                if obj not in check_list:
                    check_list.append(obj)

        if check_derives:
            _extend(osobj.get_related(Libosinfo.ProductRelationship.DERIVES_FROM).get_elements())
        if check_clones:
            _extend(osobj.get_related(Libosinfo.ProductRelationship.CLONES).get_elements())
        if check_upgrades:
            _extend(osobj.get_related(Libosinfo.ProductRelationship.UPGRADES).get_elements())

        for checkobj in check_list:
            if checkobj.get_short_id() in related_os_list or self._is_related_to(
                related_os_list,
                osobj=checkobj,
                check_upgrades=check_upgrades,
                check_derives=check_derives,
                check_clones=check_clones,
            ):
                return True

        return False

    def _get_all_devices(self):
        return list(_OsinfoIter(self._os.get_all_devices()))

    def _device_filter(self, devids=None, cls=None, extra_devs=None):
        ret = []
        devids = devids or []
        for dev in self._get_all_devices():
            if devids and dev.get_id() not in devids:
                continue
            if cls and not re.match(cls, dev.get_class()):
                continue
            ret.append(dev.get_name())

        extra_devs = extra_devs or []
        for dev in extra_devs:
            if dev.get_id() not in devids:
                continue
            ret.append(dev.get_name())

        return ret

    ###############
    # Cached APIs #
    ###############

    def _get_eol(self):
        eol = self._os.get_eol_date()
        rel = self._os.get_release_date()

        # We can use os.get_release_status() & osinfo.ReleaseStatus.ROLLING
        # if we require libosinfo >= 1.4.0.
        release_status = self._os.get_param_value(Libosinfo.OS_PROP_RELEASE_STATUS) or None

        def _glib_to_datetime(glibdate):
            date = "%s-%s" % (glibdate.get_year(), glibdate.get_day_of_year())
            return datetime.datetime.strptime(date, "%Y-%j")

        now = datetime.datetime.today()
        if eol is not None:
            return now > _glib_to_datetime(eol)

        # Rolling distributions are never EOL.
        if release_status == "rolling":
            return False

        # If no EOL is present, assume EOL if release was > 10 years ago
        if rel is not None:
            rel5 = _glib_to_datetime(rel) + datetime.timedelta(days=365 * 10)
            return now > rel5
        return False

    ###############
    # Public APIs #
    ###############

    def get_handle(self):
        return self._os

    def is_generic(self):
        return self.name == "generic"

    def is_linux_generic(self):
        return re.match(r"linux\d\d\d\d", self.name)

    def is_windows(self):
        return self._family in ["win9x", "winnt", "win16"]

    def get_clock(self):
        if self.is_windows() or self._family in ["solaris"]:
            return "localtime"
        return "utc"

    def supported_netmodels(self):
        return self._device_filter(cls="net")

    def supports_virtiodisk(self, extra_devs=None):
        # virtio-block and virtio1.0-block
        devids = ["http://pcisig.com/pci/1af4/1001", "http://pcisig.com/pci/1af4/1042"]
        return bool(self._device_filter(devids=devids, extra_devs=extra_devs))

    def supports_virtioscsi(self, extra_devs=None):
        # virtio-scsi and virtio1.0-scsi
        devids = ["http://pcisig.com/pci/1af4/1004", "http://pcisig.com/pci/1af4/1048"]
        return bool(self._device_filter(devids=devids, extra_devs=extra_devs))

    def supports_virtionet(self, extra_devs=None):
        # virtio-net and virtio1.0-net
        devids = ["http://pcisig.com/pci/1af4/1000", "http://pcisig.com/pci/1af4/1041"]
        return bool(self._device_filter(devids=devids, extra_devs=extra_devs))

    def supports_virtiorng(self, extra_devs=None):
        # virtio-rng and virtio1.0-rng
        devids = ["http://pcisig.com/pci/1af4/1005", "http://pcisig.com/pci/1af4/1044"]
        return bool(self._device_filter(devids=devids, extra_devs=extra_devs))

    def supports_virtiogpu(self, extra_devs=None):
        # virtio1.0-gpu and virtio1.0
        devids = ["http://pcisig.com/pci/1af4/1050"]
        return bool(self._device_filter(devids=devids, extra_devs=extra_devs))

    def supports_virtioballoon(self, extra_devs=None):
        # virtio-balloon and virtio1.0-balloon
        devids = ["http://pcisig.com/pci/1af4/1002", "http://pcisig.com/pci/1af4/1045"]
        return bool(self._device_filter(devids=devids, extra_devs=extra_devs))

    def supports_virtioserial(self, extra_devs=None):
        devids = ["http://pcisig.com/pci/1af4/1003", "http://pcisig.com/pci/1af4/1043"]
        if self._device_filter(devids=devids, extra_devs=extra_devs):
            return True
        # osinfo data was wrong for RHEL/centos here until Oct 2018
        # Remove this hack after 6 months or so
        return self._is_related_to("rhel6.0")

    def supports_virtioinput(self, extra_devs=None):
        # virtio1.0-input
        devids = ["http://pcisig.com/pci/1af4/1052"]
        return bool(self._device_filter(devids=devids, extra_devs=extra_devs))

    def supports_usb3(self, extra_devs=None):
        # qemu-xhci
        devids = ["http://pcisig.com/pci/1b36/0004"]
        return bool(self._device_filter(devids=devids, extra_devs=extra_devs))

    def supports_virtio1(self, extra_devs=None):
        # Use virtio1.0-net device as a proxy for virtio1.0 as a whole
        devids = ["http://pcisig.com/pci/1af4/1041"]
        return bool(self._device_filter(devids=devids, extra_devs=extra_devs))

    def supports_chipset_q35(self, extra_devs=None):
        # For our purposes, check for the union of q35 + virtio1.0 support
        if self.supports_virtionet(extra_devs=extra_devs) and not self.supports_virtio1(
            extra_devs=extra_devs
        ):
            return False
        devids = ["http://qemu.org/chipset/x86/q35"]
        return bool(self._device_filter(devids=devids, extra_devs=extra_devs))

    def _get_firmware_list(self):
        if hasattr(self._os, "get_complete_firmware_list"):  # pragma: no cover
            return self._os.get_complete_firmware_list().get_elements()
        return []  # pragma: no cover

    def _supports_firmware_type(self, name, arch, default):
        firmwares = self._get_firmware_list()

        for firmware in firmwares:  # pragma: no cover
            if firmware.get_architecture() != arch:
                continue
            if firmware.get_firmware_type() == name:
                return firmware.is_supported()

        return default

    def requires_firmware_efi(self, arch):
        ret = False
        try:
            supports_efi = self._supports_firmware_type("efi", arch, False)
            supports_bios = self._supports_firmware_type("bios", arch, True)
            ret = supports_efi and not supports_bios
        except Exception:  # pragma: no cover
            log.debug("Error checking osinfo firmware support", exc_info=True)

        return ret

    def get_recommended_resources(self):
        minimum = self._os.get_minimum_resources()
        recommended = self._os.get_recommended_resources()
        return _OsResources(minimum, recommended)

    def get_network_install_required_ram(self, guest):
        if hasattr(self._os, "get_network_install_resources"):
            resources = self._os.get_network_install_resources()
            for r in _OsinfoIter(resources):
                arch = r.get_architecture()
                if arch == guest.os.arch or arch == "all":
                    return r.get_ram()

    def get_kernel_url_arg(self):
        """
        Kernel argument name the distro's installer uses to reference
        a network source, possibly bypassing some installer prompts
        """
        # Let's ask the OS for its kernel argument for the source
        if hasattr(self._os, "get_kernel_url_argument"):
            osarg = self._os.get_kernel_url_argument()
            if osarg is not None:
                return osarg

        # SUSE distros
        if self.distro in ["caasp", "sle", "sled", "sles", "opensuse"]:
            return "install"

        if self.distro not in ["centos", "rhel", "fedora"]:
            return None

        # Default for RH distros, in case libosinfo data isn't complete
        return "inst.repo"  # pragma: no cover

    def _get_generic_location(self, treelist, arch, profile):
        if not hasattr(Libosinfo.Tree, "get_os_variants"):  # pragma: no cover
            for tree in treelist:
                if tree.get_architecture() == arch:
                    return tree.get_url()
            return None

        fallback_tree = None
        if profile == "jeos":
            profile = "Server"
        elif profile == "desktop":
            profile = "Workstation"
        elif not profile:
            profile = "Everything"

        for tree in treelist:
            if tree.get_architecture() != arch:
                continue

            variant_list = tree.get_os_variants()
            fallback_tree = tree
            for variant in _OsinfoIter(variant_list):
                if profile in variant.get_name():
                    return tree.get_url()

        if fallback_tree:
            return fallback_tree.get_url()
        return None

    def get_location(self, arch, profile=None):
        treelist = list(_OsinfoIter(self._os.get_tree_list()))

        if not treelist:
            raise RuntimeError(_("OS '%s' does not have a URL location") % self.name)

        # Some distros have more than one URL for a specific architecture,
        # which is the case for Fedora and different variants (Server,
        # Workstation). Later on, we'll have to differentiate that and return
        # the right one. However, for now, let's just rely on returning the
        # most generic tree possible.
        location = self._get_generic_location(treelist, arch, profile)
        if location:
            return location

        raise RuntimeError(
            _("OS '%(osname)s' does not have a URL location for the architecture '%(archname)s'")
            % {"osname": self.name, "archname": arch}
        )

    def get_install_script_list(self):
        return list(_OsinfoIter(self._os.get_install_script_list()))

    def _get_installable_drivers(self, arch):
        installable_drivers = []
        device_drivers = list(_OsinfoIter(self._os.get_device_drivers()))
        for device_driver in device_drivers:
            if arch != "all" and device_driver.get_architecture() != arch:
                continue

            installable_drivers.append(device_driver)
        return installable_drivers

    def _get_pre_installable_drivers(self, arch):
        installable_drivers = self._get_installable_drivers(arch)
        pre_inst_drivers = []
        for driver in installable_drivers:
            if driver.get_pre_installable():
                pre_inst_drivers.append(driver)
        return pre_inst_drivers

    def _get_drivers_location(self, drivers):
        locations = []
        for driver in drivers:
            filenames = driver.get_files()
            for filename in filenames:
                location = os.path.join(driver.get_location(), filename)
                locations.append(location)
        return locations

    def get_pre_installable_drivers_location(self, arch):
        pre_inst_drivers = self._get_pre_installable_drivers(arch)

        return self._get_drivers_location(pre_inst_drivers)

    def get_pre_installable_devices(self, arch):
        drivers = self._get_pre_installable_drivers(arch)
        devices = []
        for driver in drivers:
            devices += list(_OsinfoIter(driver.get_devices()))
        return devices

    def supports_unattended_drivers(self, arch):
        if self._get_pre_installable_drivers(arch):
            return True
        return False


class _OsMedia:
    def __init__(self, osinfo_media):
        self._media = osinfo_media

    def get_kernel_path(self):
        return self._media.get_kernel_path()

    def get_initrd_path(self):
        return self._media.get_initrd_path()

    def supports_installer_script(self):
        return self._media.supports_installer_script()

    def is_netinst(self):
        variants = list(_OsinfoIter(self._media.get_os_variants()))
        for variant in variants:
            if "netinst" in variant.get_id():
                return True
        return False  # pragma: no cover

    def get_install_script_list(self):
        return list(_OsinfoIter(self._media.get_install_script_list()))

    def get_osinfo_media(self):
        return self._media


class _OsTree:
    def __init__(self, osinfo_tree):
        self._tree = osinfo_tree

    def get_osinfo_tree(self):
        return self._tree



# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

import os
import shutil

import tests.utils
from . import lib


class _DeleteRow:
    """
    Helper class for interacting with the delete dialog rows
    """

    def __init__(self, cell1, cell2, cell3, cell4):
        ignore = cell4
        self.chkcell = cell1
        self.path = cell2.text
        self.target = cell3.text
        self.undeletable = not self.chkcell.sensitive
        self.default = self.chkcell.checked
        self.notdefault = not self.undeletable and not self.default


def _create_testdriver_path(fn):
    def wrapper(app, *args, **kwargs):
        # This special path is hardcoded in test-many-devices
        tmppath = "/tmp/virt-manager-uitests/tmp1"
        tmpdir = os.path.dirname(tmppath)
        try:
            if not os.path.exists(tmpdir):
                os.mkdir(tmpdir)
            open(tmppath, "w").write("foo")
            os.chmod(tmppath, 0o444)
            return fn(app, tmppath, *args, **kwargs)
        finally:
            if os.path.exists(tmpdir):
                os.chmod(tmpdir, 0o777)
                shutil.rmtree(tmpdir)

    return wrapper


def _open_storage_browser(app):
    app.root.find("New", "push button").click()
    newvm = app.find_window("New VM")
    newvm.find_fuzzy("Local install media", "radio").click()
    newvm.find_fuzzy("Forward", "button").click()
    newvm.find_fuzzy("install-iso-browse", "button").click()
    return app.root.find("vmm-storage-browser")


def _open_delete(app, vmname):
    app.manager_vm_action(vmname, delete=True)
    return app.find_window("Delete")


def _finish(app, delete, paths, expect_fail=False, click_no=False):
    delete.find_fuzzy("Delete", "button").click()
    if paths:
        alert = app.root.find("vmm dialog", "alert")
        alert.find_fuzzy("Are you sure")
        for path in paths:
            alert.find_fuzzy(path)
        if click_no:
            alert.find("No", "push button").click()
            return
        alert.find("Yes", "push button").click()
    if not expect_fail:
        lib.utils.check(lambda: not delete.showing)


def _get_all_rows(delete):
    slist = delete.find("storage-list")

    def pred(node):
        return node.roleName == "table cell"

    cells = slist.findChildren(pred, isLambda=True)

    idx = 0
    rows = []
    while idx < len(cells):
        rows.append(_DeleteRow(*cells[idx : idx + 4]))
        idx += 4
    return rows


################################################
# UI tests for virt-manager's VM delete window #
################################################


def _testDeleteManyDevices(app, nondefault_path=None, delete_nondefault=False, skip_finish=False):
    delete = _open_delete(app, "test-many-devices")

    rows = _get_all_rows(delete)
    selected_rows = [r.path for r in rows if r.default]
    undeletable_rows = [r.path for r in rows if r.undeletable]
    notdefault_rows = [r.path for r in rows if r.notdefault]

    defpath = "/pool-dir/overlay.img"
    nondefault_path2 = "/pool-dir/sharevol.img"

    assert selected_rows == [defpath]
    if nondefault_path:
        assert nondefault_path in notdefault_rows
    assert nondefault_path2 in notdefault_rows
    assert "/dev/fda" in undeletable_rows

    if delete_nondefault:
        # Click the selector for the nondefault path
        found = [r for r in rows if r.path == nondefault_path]
        assert len(found) == 1
        slist = delete.find("storage-list")
        slist.click()
        chkcell = found[0].chkcell
        chkcell.bring_on_screen()
        chkcell.click()
        chkcell.click()
        chkcell.click()
        lib.utils.check(lambda: chkcell.checked)

    paths = []
    if defpath:
        paths.append(defpath)
    if delete_nondefault:
        paths.append(nondefault_path)
    if skip_finish:
        return paths
    _finish(app, delete, paths)

    # Confirm
    browser = _open_storage_browser(app)
    browser.find_fuzzy("pool-dir", "table cell").click()
    browser.find("vol-refresh", "push button").click()
    lib.utils.check(lambda: "overlay.img" not in browser.fmt_nodes())
    browser.find("sharevol.img", "table cell")


@_create_testdriver_path
def testDeleteManyDevices(app, tmppath):
    """
    Hit a specific case of a path not selected by default
    because the permissions are readonly
    """
    _testDeleteManyDevices(app, nondefault_path=tmppath)


@_create_testdriver_path
def testDeleteNondefaultOverride(app, tmppath):
    """
    Path not selected by default, but we select it,
    which will cause it to be manually unlinked
    """
    _testDeleteManyDevices(app, nondefault_path=tmppath, delete_nondefault=True)
    assert not os.path.exists(tmppath)


@_create_testdriver_path
def testDeleteFailure(app, tmppath):
    """
    After launching the wizard we change permissions to make
    file deletion fail
    """
    paths = _testDeleteManyDevices(
        app, nondefault_path=tmppath, delete_nondefault=True, skip_finish=True
    )
    os.chmod(os.path.dirname(tmppath), 0o555)
    delete = app.find_window("Delete")
    _finish(app, delete, paths, expect_fail=True, click_no=True)
    lib.utils.check(lambda: delete.active)
    _finish(app, delete, paths, expect_fail=True)
    assert os.path.exists(tmppath)
    app.click_alert_button("Errors encountered", "Close")

    # Ensure disconnecting will close the dialog
    win = _open_delete(app, "test-clone")
    app.manager_test_conn_window_cleanup("test testdriver.xml", win)


def testDeleteRemoteManyDevices(app):
    """
    Test with a remote VM to hit a certain code path
    """
    app.uri = tests.utils.URIs.kvm_x86_remote
    _testDeleteManyDevices(app)


def testDeleteSkipStorage(app):
    """
    Test VM delete with all storage skipped
    """
    delete = _open_delete(app, "test-many-devices")
    chk = delete.find("Delete associated", "check box")
    slist = delete.find("storage-list")

    lib.utils.check(lambda: chk.checked)
    chk.click()
    lib.utils.check(lambda: not chk.checked)
    lib.utils.check(lambda: not slist.showing)

    _finish(app, delete, None)

    # Confirm nothing was deleted compare to the default selections
    browser = _open_storage_browser(app)
    browser.find_fuzzy("pool-dir", "table cell").click()
    browser.find("vol-refresh", "push button").click()
    browser.find("overlay.img", "table cell")
    browser.find("sharevol.img", "table cell")


def testDeleteDeviceNoStorage(app):
    """
    Verify successful device remove with storage doesn't
    touch host storage
    """
    details = app.manager_open_details("test-many-devices", shutdown=True)

    hwlist = details.find("hw-list")
    hwlist.click()
    c = hwlist.find("USB Disk 1")
    c.bring_on_screen()
    c.click()
    tab = details.find("disk-tab")
    lib.utils.check(lambda: tab.showing)
    details.find("config-remove").click()

    delete = app.find_window("Remove Disk")
    chk = delete.find("Delete associated", "check box")
    lib.utils.check(lambda: not chk.checked)
    _finish(app, delete, [])
    details.window_close()

    browser = _open_storage_browser(app)
    browser.find_fuzzy("pool-dir", "table cell").click()
    browser.find("vol-refresh", "push button").click()
    browser.find("overlay.img", "table cell")


def testDeleteDeviceWithStorage(app):
    """
    Verify successful device remove deletes storage
    """
    details = app.manager_open_details("test-many-devices", shutdown=True)

    hwlist = details.find("hw-list")
    hwlist.click()
    c = hwlist.find("USB Disk 1")
    c.bring_on_screen()
    c.click()
    tab = details.find("disk-tab")
    lib.utils.check(lambda: tab.showing)
    details.find("config-remove").click()

    delete = app.find_window("Remove Disk")
    chk = delete.find("Delete associated", "check box")
    lib.utils.check(lambda: not chk.checked)
    chk.click()
    lib.utils.check(lambda: chk.checked)
    path = "/pool-dir/overlay.img"
    delete.find_fuzzy(path)
    _finish(app, delete, [path])
    details.window_close()

    browser = _open_storage_browser(app)
    browser.find_fuzzy("pool-dir", "table cell").click()
    browser.find("vol-refresh", "push button").click()
    lib.utils.check(lambda: "overlay.img" not in browser.fmt_nodes())


def testDeleteDeviceFail(app):
    """
    Verify failed device remove does not touch storage
    """
    details = app.manager_open_details("test-many-devices")

    hwlist = details.find("hw-list")
    hwlist.click()
    c = hwlist.find("USB Disk 1")
    c.bring_on_screen()
    c.click()
    tab = details.find("disk-tab")
    lib.utils.check(lambda: tab.showing)
    details.find("config-remove").click()

    delete = app.find_window("Remove Disk")
    chk = delete.find("Delete associated", "check box")
    lib.utils.check(lambda: not chk.checked)
    chk.click()
    lib.utils.check(lambda: chk.checked)
    path = "/pool-dir/overlay.img"
    delete.find_fuzzy(path)
    _finish(app, delete, [path], expect_fail=True)
    app.click_alert_button("Storage will not be.*deleted", "OK")
    details.window_close()

    # Verify file still exists
    browser = _open_storage_browser(app)
    browser.find_fuzzy("pool-dir", "table cell").click()
    browser.find("vol-refresh", "push button").click()
    browser.find("overlay.img", "table cell")



#
# Copyright 2009, 2013 Red Hat, Inc.
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

from .device import Device
from ..xmlbuilder import XMLProperty


class DeviceInput(Device):
    XML_NAME = "input"

    TYPE_MOUSE = "mouse"
    TYPE_TABLET = "tablet"
    TYPE_KEYBOARD = "keyboard"
    TYPE_EVDEV = "evdev"

    BUS_PS2 = "ps2"
    BUS_USB = "usb"
    BUS_VIRTIO = "virtio"
    BUS_XEN = "xen"

    type = XMLProperty("./@type")
    bus = XMLProperty("./@bus")
    model = XMLProperty("./@model")

    source_evdev = XMLProperty("./source/@evdev")
    source_dev = XMLProperty("./source/@dev")
    source_repeat = XMLProperty("./source/@repeat", is_onoff=True)
    source_grab = XMLProperty("./source/@grab")
    source_grabToggle = XMLProperty("./source/@grabToggle")

    ##################
    # Default config #
    ##################

    def set_defaults(self, guest):
        if not self.type:
            self.type = self.TYPE_MOUSE



# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

from . import lib


#####################################
# UI tests for the createnet wizard #
#####################################


def _open_netadd(app, hostwin):
    hostwin.find("net-add", "push button").click()
    win = app.find_window("Create a new virtual network")
    return win


def testCreateNet(app):
    """
    Basic test with object state management afterwards
    """
    hostwin = app.manager_open_host("Virtual Networks")
    win = _open_netadd(app, hostwin)

    # Create a simple default network
    name = win.find("Name:", "text")
    finish = win.find("Finish", "push button")
    lib.utils.check(lambda: name.text == "network")
    newname = "a-test-new-net"
    name.set_text(newname)
    finish.click()

    # Select the new network in the host window, then do
    # stop->start->stop->delete, for lifecycle testing
    lib.utils.check(lambda: hostwin.active)
    cell = hostwin.find(newname, "table cell")
    delete = hostwin.find("net-delete", "push button")
    start = hostwin.find("net-start", "push button")
    stop = hostwin.find("net-stop", "push button")

    cell.click()
    stop.click()
    lib.utils.check(lambda: start.sensitive)
    start.click()
    lib.utils.check(lambda: stop.sensitive)
    stop.click()
    lib.utils.check(lambda: delete.sensitive)

    # Delete it, clicking No first
    delete.click()
    app.click_alert_button("permanently delete the network", "No")
    lib.utils.check(lambda: not cell.dead)
    delete.click()
    app.click_alert_button("permanently delete the network", "Yes")
    # Ensure it's gone
    lib.utils.check(lambda: cell.dead)


def testCreateNetXMLEditor(app):
    """
    Test the XML editor
    """
    app.open(xmleditor_enabled=True)
    hostwin = app.manager_open_host("Virtual Networks")
    win = _open_netadd(app, hostwin)
    name = win.find("Name:", "text")
    finish = win.find("Finish", "push button")

    # Create a new obj with XML edited name, verify it worked
    tmpname = "objtmpname"
    newname = "froofroo"
    name.set_text(tmpname)
    win.find("XML", "page tab").click()
    xmleditor = win.find("XML editor")
    newtext = xmleditor.text.replace(">%s<" % tmpname, ">%s<" % newname)
    xmleditor.set_text(newtext)
    finish.click()
    lib.utils.check(lambda: hostwin.active)
    cell = hostwin.find(newname, "table cell")
    cell.click()

    # Do standard xmleditor tests
    win = _open_netadd(app, hostwin)
    lib.utils.test_xmleditor_interactions(app, win, finish)
    win.find("Cancel", "push button").click()
    lib.utils.check(lambda: not win.visible)

    # Ensure host window closes fine
    hostwin.click()
    hostwin.keyCombo("<ctrl>w")
    lib.utils.check(lambda: not hostwin.showing and not hostwin.active)


def testCreateNetMulti(app):
    """
    Test remaining create options
    """
    app.uri = "test:///default"
    hostwin = app.manager_open_host("Virtual Networks", conn_label="test default")
    win = _open_netadd(app, hostwin)
    finish = win.find("Finish", "push button")

    # Create a network with a bunch of options
    win.find("Name:", "text").set_text("default")
    win.find("net-mode").click()
    win.find("Isolated", "menu item").click()
    win.find("IPv4 configuration").click_expander()
    win.find("ipv4-network").set_text("192.168.100.0/25")
    ipv4start = win.find("ipv4-start")
    ipv4end = win.find("ipv4-end")
    lib.utils.check(lambda: ipv4start.text == "192.168.100.64")
    lib.utils.check(lambda: ipv4end.text == "192.168.100.126")
    win.find("Enable DHCPv4").click()
    win.find("Enable IPv4").click()
    win.find("IPv6 configuration").click_expander()
    win.find("Enable IPv6").click()
    win.find("Enable DHCPv6").click()
    win.find("ipv6-network").set_text("fd00:beef:10:6::1/64")
    win.find("ipv6-start").set_text("fd00:beef:10:6::1:1")
    win.find("ipv6-end").set_text("bad")
    win.find("DNS domain name").click_expander()
    win.find("Custom").click()
    win.find("domain-custom").set_text("mydomain")
    finish.click()
    # Name collision validation
    app.click_alert_button("in use by another network", "Close")
    win.find("Name:", "text").set_text("newnet1")
    finish.click()
    # XML define error
    app.click_alert_button("Error creating virtual network", "Close")
    win.find("ipv6-end").set_text("fd00:beef:10:6::1:f1")
    finish.click()
    lib.utils.check(lambda: hostwin.active)

    # More option work
    win = _open_netadd(app, hostwin)
    win.find("Name:", "text").set_text("newnet2")
    devicelist = win.find("net-devicelist")
    lib.utils.check(lambda: not devicelist.visible)
    win.find("net-mode").click()
    win.find("SR-IOV", "menu item").click()
    lib.utils.check(lambda: devicelist.visible)
    # Just confirm this is here
    win.find("No available device", "menu item")
    win.find("net-mode").click()
    win.find("Routed", "menu item").click()
    win.find("net-forward").click()
    win.find("Physical device", "menu item").click()
    win.find("net-device").set_text("fakedev0")
    finish.click()
    lib.utils.check(lambda: hostwin.active)


def testCreateNetSRIOV(app):
    """
    We need the full URI to test the SRIOV method
    """
    app.open(xmleditor_enabled=True)
    hostwin = app.manager_open_host("Virtual Networks")
    win = _open_netadd(app, hostwin)
    finish = win.find("Finish", "push button")

    win.find("net-mode").click()
    win.find("SR-IOV", "menu item").click()
    win.find("net-devicelist").click()
    win.find_fuzzy("eth3", "menu item").click()
    finish.click()



# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

from . import lib


#############################################
# UI tests for the 'open connection' dialog #
#############################################


def testConnect(app):
    # Start with connection delete
    c = app.manager_conn_disconnect("test testdriver.xml")
    c.click(button=3)
    app.root.find("conn-delete", "menu item").click()
    app.click_alert_button("will remove the connection", "No")
    lib.utils.check(lambda: not c.dead)
    app.manager_conn_delete("test testdriver.xml")

    # Launch the dialog, grab some UI pointers
    win = app.manager_open_createconn()
    connect = win.find("Connect", "push button")
    remote = win.find_fuzzy("Connect to remote", "check box")
    user = win.find("Username", "text")
    host = win.find("Hostname", "text")
    urilabel = win.find("uri-label", "label")
    lib.utils.check(lambda: user.showing is host.showing is True)

    # Select all HV options
    win.combo_select("Hypervisor", "QEMU/KVM user session")
    win.combo_select("Hypervisor", r"^QEMU/KVM$")
    win.combo_select("Hypervisor", "Xen")
    win.combo_select("Hypervisor", "Bhyve")
    win.combo_select("Hypervisor", "Virtuozzo")
    win.combo_select("Hypervisor", r".*LXC.*")

    # Test a simple selection
    win.combo_select("Hypervisor", "QEMU/KVM user session")
    lib.utils.check(lambda: user.showing is host.showing is False)
    lib.utils.check(lambda: urilabel.text == "qemu:///session")

    # Cancel the dialog
    win.find_fuzzy("Cancel", "push button").click()
    lib.utils.check(lambda: not win.showing)

    # Reopen it, confirm content changed
    win = app.manager_open_createconn()
    lib.utils.check(lambda: ":///session" not in urilabel.text)

    # Relaunch the dialog, confirm it doesn't overwrite content
    win.combo_select("Hypervisor", ".*LXC.*")
    lib.utils.check(lambda: "lxc" in urilabel.text)
    win = app.manager_open_createconn()
    lib.utils.check(lambda: win.active)
    lib.utils.check(lambda: "lxc" in urilabel.text)

    # Enter a failing URI, make sure error is raised, and we can
    # fall back to the dialog
    win.combo_select("Hypervisor", "Xen")
    remote.click()
    user.set_text("fribuser")
    connect.click()
    app.click_alert_button("hostname is required", "OK")
    fakeipv6 = "fe80::1"
    host.set_text(fakeipv6)
    lib.utils.check(lambda: urilabel.text == "xen+ssh://fribuser@[%s]/" % fakeipv6)
    fakehost = "ix8khfyidontexistkdjur.com"
    host.set_text(fakehost + ":12345")
    lib.utils.check(lambda: urilabel.text == "xen+ssh://fribuser@%s:12345/" % fakehost)
    connect.click()

    lib.utils.check(lambda: win.showing is True)
    c = app.root.find_fuzzy(fakehost, "table cell")
    lib.utils.check(lambda: "Connecting..." not in c.text, timeout=10)
    app.click_alert_button("Unable to connect", "No")

    # Ensure dialog shows old contents for editing
    lib.utils.check(lambda: win.showing)
    lib.utils.check(lambda: fakehost in host.text)

    # This time say 'yes'
    connect.click()
    lib.utils.check(lambda: win.showing is True)
    c = app.root.find_fuzzy(fakehost, "table cell")
    lib.utils.check(lambda: "Connecting..." not in c.text, timeout=10)
    app.click_alert_button("Unable to connect", "Yes")
    c = app.root.find_fuzzy(fakehost, "table cell")
    lib.utils.check(lambda: win.showing is False)

    # Test with custom test:///default connection
    app.manager_createconn("test:///default")
    # Do it again to make sure things don't explode
    app.manager_createconn("test:///default")

    # Test connection double click
    c = app.manager_conn_disconnect("test default")
    c.doubleClick()
    lib.utils.check(lambda: "Not Connected" not in c.text)

    # Delete it
    app.manager_conn_disconnect("test default")
    app.manager_conn_delete("test default")



# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

import os

import pytest

from virtinst import cli
from virtinst import pollhelpers
from virtinst import StoragePool
from virtinst import URI


############################
# VirtinstConnection tests #
############################


def test_misc():
    # Misc API checks
    conn = cli.getConnection("test:///default")
    conn.invalidate_caps()
    assert conn.is_open() is True
    assert conn.is_container_only() is False
    assert conn.is_openvz() is False
    assert not conn.get_uri_hostname()
    assert not conn.get_uri_port()
    assert not conn.get_uri_username()
    assert not conn.get_uri_transport()
    assert conn.close() == 0

    # Coverage for a daemon_version check
    fakeuri = "__virtinst_test__test:///default,libver=123"
    conn = cli.getConnection(fakeuri)
    assert conn.daemon_version() == 123

    # Hit a special code path that reflects default libvirt transport
    # pylint: disable=protected-access
    conn._uriobj = URI("qemu://example.com/system")
    assert conn.get_uri_transport() == "tls"

    # Hit the qemu:///embed case privileged case check
    fakeuri = "__virtinst_test__test:///default,fakeuri=qemu:///embed"
    conn = cli.getConnection(fakeuri)
    assert conn.is_privileged() == (os.getuid() == 0)

    # Hit fakuuri validation error, for old style opts
    with pytest.raises(RuntimeError):
        cli.getConnection(fakeuri + ",qemu")


def test_default_uri(monkeypatch):
    monkeypatch.setitem(os.environ, "LIBVIRT_DEFAULT_URI", "test:///default")

    # Handle connecting to None conn
    conn = cli.getConnection(None)
    assert conn.getURI() == "test:///default"
    conn.close()


def test_poll():
    # Add coverage for conn fetch_* handling, and pollhelpers
    conn = cli.getConnection("test:///default")
    objmap = {}

    def build_cb(obj, name):
        return obj

    gone, new, master = pollhelpers.fetch_nets(conn, {}, build_cb)
    assert len(gone) == 0
    assert len(new) == 1
    assert len(master) == 1
    assert master[0].name() == "default"

    objmap = dict((obj.name(), obj) for obj in master)
    gone, new, master = pollhelpers.fetch_nets(conn, objmap, build_cb)
    assert len(gone) == 0
    assert len(new) == 0
    assert len(master) == 1
    assert master[0].name() == "default"

    # coverage for some special cases in cache_new_pool
    def makepool(name, create):
        poolxml = StoragePool(conn)
        poolxml.type = "dir"
        poolxml.name = name
        poolxml.target_path = "/tmp/foo/bar/baz/%s" % name
        return poolxml.install(create=create)

    poolobj1 = makepool("conntest1", False)
    conn.fetch_all_pools()
    poolobj2 = makepool("conntest2", True)
    conn.fetch_all_vols()
    poolobj1.undefine()
    poolobj2.destroy()
    poolobj2.undefine()
